from colorama import Fore
import streamanalyser as sa


# Customize intensity options
if __name__ == "__main__":
    analyser_obj = sa.StreamAnalyser(
        "ShB4Wen_HBg",
        verbose=True,
        not_cache=True,
        disable_logs=True,
        msglimit=5000,
        intensity_levels=["LOW", "MEDIUM", "HIGH"],
        intensity_constants=[0, 1, 2],
        intensity_colors=[Fore.YELLOW, Fore.RED, Fore.MAGENTA],
    )
    with analyser_obj as analyser:
        analyser.analyse()
        analyser.get_highlights(output_mode="summary")



from . import utils
from . import loggersetup
from .structures import Emote, Icon, Membership, Message, Author, Money, Sticker, Superchat, SuperchatColor


class DataRefiner:
    """Refines raw data into a usable form"""

    def __init__(self, log_path=None, verbose=False):
        self.verbose = verbose

        self.messages = []
        self.authors = []
        self.logger = loggersetup.create_logger(__file__, log_path)

    def refine_raw_messages(self, raw_messages, msglimit=None) -> list[Message,Superchat,Membership]:
        """Refines raw messages and shapes them into Message dataclass.
            
            Also gets all unique authors. This behavior was separate as per single responsibility principle
            but now they're merged to improve performance. 
        """

        self.logger.info("Refining messages")
        messages = []
        authors = []
        skipped_message_amount = 0
        if self.verbose:
            print(f"Refining messagess...0%", end="\r")
        for count, raw_message in enumerate(raw_messages):
            if msglimit and count == msglimit:
                break
            if self.verbose:
                print(
                    f"Refining messages...{utils.percentage(count, msglimit if msglimit else len(raw_messages))}%",
                    end="\r",
                )
            try:
                convertedMessage = self._convert_message(raw_message)
                messages.append(convertedMessage)
                authors.append(convertedMessage.author)
            except ValueError as e:
                self.logger.warning(f"{e.__class__.__name__}: {e}")
                self.logger.debug(f"Corrupt message was {raw_message}")
                skipped_message_amount += 1
            except Exception as e:
                self.logger.error(f"{e.__class__.__name__}:{e}")
                skipped_message_amount += 1
        self.logger.debug(f"{len(messages)} messages has been refined ({skipped_message_amount} skipped)")
        self.logger.debug(f"{len(self.authors)} authors has been found")
        if self.verbose:
            print(f"Refining messages... done")
        self.messages = messages
        self.authors = list(dict.fromkeys(authors))
        return messages

    def _convert_message(self, raw_message):
        """Converts raw message to data classes"""

        text = raw_message["message"]
        if not text:
            text = ""
        emotes=[]
        if "emotes" in raw_message.keys():
            emotes=[Emote(
                id=emote["id"],
                name=emote["name"],
                is_custom_emoji=emote["is_custom_emoji"],
                images=[Icon(
                    id=img["id"] if emote["is_custom_emoji"] else "None",
                    url=img["url"],
                    height=img["height"] if "height" in img.keys() else 0,
                    width=img["width"] if "width" in img.keys() else 0,
                ) for img in emote["images"]],
            ) for emote in raw_message["emotes"]]

        is_member=False
        membership_info=""
        membership_icons=[]
        if "badges" in raw_message["author"].keys():
            if "icons" in raw_message["author"]["badges"][0].keys():
                membership_icons = [
                        Icon(
                            id=img["id"],
                            url=img["url"],
                            height=img["height"] if "height" in img.keys() else 0,
                            width=img["width"] if "width" in img.keys() else 0,
                        ) for img in raw_message["author"]["badges"][0]["icons"]]
            for badge in raw_message["author"]["badges"]:
                if "member" in badge["title"].lower():
                    is_member=True
                    membership_info=badge["title"]

        author=Author(
            id=raw_message["author"]["id"],
            name=raw_message["author"]["name"],
            is_member=is_member,
            membership_info=membership_info,
            images={
                "profile":[
                    Icon(
                        id=img["id"],
                        url=img["url"],
                        height=img["height"] if "height" in img.keys() else 0,
                        width=img["width"] if "width" in img.keys() else 0,
                    ) for img in raw_message["author"]["images"]
                ],
                "membership": membership_icons
            }
        )
        if raw_message.get("message_type") == "text_message":
            return Message(
                id=raw_message["message_id"],
                text=text,
                time=round(raw_message["time_in_seconds"]),
                author=author,
                emotes=emotes
            )
        elif raw_message.get("message_type") == "paid_message":
            return Superchat(
                id=raw_message["message_id"],
                text=text,
                time=round(raw_message["time_in_seconds"]),
                author=author,
                money=Money(
                    amount=raw_message["money"]["amount"],
                    currency=raw_message["money"]["currency"],
                    currency_symbol=raw_message["money"]["currency_symbol"],
                    text=raw_message["money"]["text"],
                ),
                colors=SuperchatColor(
                    background=raw_message["colors"]["body_background_colour"],
                    header=raw_message["colors"]["header_background_colour"]
                ),
                emotes=emotes
            )
        elif raw_message.get("message_type") == "membership_item":
            return Membership(
                id=raw_message["message_id"],
                text=text,
                welcome_text=raw_message["welcome_text"],
                time=round(raw_message["time_in_seconds"]),
                author=author,
                emotes=emotes
            )
        elif raw_message.get("message_type") == "paid_sticker":
            return Sticker(
                id=raw_message["message_id"],
                text=text,
                time=round(raw_message["time_in_seconds"]),
                author=author,
                money=Money(
                    amount=raw_message["money"]["amount"],
                    currency=raw_message["money"]["currency"],
                    currency_symbol=raw_message["money"]["currency_symbol"],
                    text=raw_message["money"]["text"],
                ),
                colors=SuperchatColor(
                    background=raw_message["colors"]["body_background_colour"],
                    header=raw_message["colors"]["header_background_colour"]
                ),
                emotes=emotes,
                sticker_images=raw_message["sticker_images"],
            )
        raise ValueError(f"Invalid message type: {raw_message.get('message_type')}")


    def get_authors(self) -> list[Author]:
        """Returns unique list of message authors"""

        if not self.authors:
            self.logger.warning("Raw messages should be refined before getting authors")
            return []

        return self.authors



import json
import os
import unittest
import random
import warnings

from modules.chatanalyser import ChatAnalyser, Fore
from modules.structures import (
    Intensity,
    Message,
    Author,
    Context,
    Trigger
)
from modules.exceptions import (
    DifferentListSizeError,
    ConstantsNotAscendingError,
    ConstantsNotUniqueError,
    DuplicateContextException,
    UnexpectedException,
    ContextsAllCorruptException,
    PathAlreadyExistsException
)


def generate_random_chat(size, seed, density=0):
    random.seed(seed)
    sample_messages = []
    i = 0
    while i < size:
        if not i == 0 and random.choice(list(range(density))):
            i -= 1
        i += 1
        rid = random.random()
        sample_messages.append(
            Message(
                id=rid,
                text="msg" + str(i),
                time=i,
                author=Author(id="u" + str(rid), name="usr" + str(i)),
            )
        )

    return sample_messages

SAMPLE_CONTEXT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "contexts.json"
)
ANOTHER_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "another_context.json"
)
PARTIAL_CORRUPT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "partial_corrupt_context.json"
)
FULL_CORRUPT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "full_corrupt_context.json"
)

sample_contexts = [
            {
                "reaction_to": "reaction1",
                "triggers": [
                    {"phrase": "msg6", "is_exact": False},
                    {"phrase": "msg34", "is_exact": False},
                    {"phrase": "msg2", "is_exact": False},
                    {"phrase": "msg32", "is_exact": False},
                ],
            },
            {
                "reaction_to": "reaction2",
                "triggers": [
                    {"phrase": "msg42msg40", "is_exact": True},
                    {"phrase": "msg49", "is_exact": True},
                    {"phrase": "msg40", "is_exact": True},
                ],
            },
            {
                "reaction_to": "reaction3",
                "triggers": [
                    {"phrase": "msg74", "is_exact": False},
                    {"phrase": "msg35", "is_exact": False},
                    {"phrase": "msg77", "is_exact": False},
                ],
            },
        ]

partially_corrupt_contexts = [
    {"wrong_key": "reaction"},
    {
        "reaction_to": "reaction3",
        "triggers": [
            {"phrase": "msg35", "is_exact": False},
        ],
    },
]

fully_corrupt_contexts = [
    {"wrong_key": "reaction"},
    {"wrong_key2": "reaction"},
]


class TestChatAnalyser(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", category=ResourceWarning)

        with open(SAMPLE_CONTEXT_PATH, "w") as file:
            file.write(json.dumps(sample_contexts))
        with open(ANOTHER_PATH, "w") as file:
            file.write(json.dumps(sample_contexts))
        with open(PARTIAL_CORRUPT_PATH, "w") as file:
            file.write(json.dumps(partially_corrupt_contexts))
        with open(FULL_CORRUPT_PATH, "w") as file:
            file.write(json.dumps(fully_corrupt_contexts))
        self.canalyser = ChatAnalyser(
            generate_random_chat(20, 101, 3),
            log_path=None,
            default_context_path=None
        )
        self.canalyser.logger.disabled = True

    def tearDown(self):
        os.remove(SAMPLE_CONTEXT_PATH)
        os.remove(ANOTHER_PATH)
        os.remove(PARTIAL_CORRUPT_PATH)
        os.remove(FULL_CORRUPT_PATH)
        del self.canalyser

    def test_get_frequency(self):
        result = self.canalyser.get_frequency()
        expected = {
            0: 0,
            1: 1,
            2: 4,
            3: 2,
            4: 2,
            5: 5,
            6: 6,
            7: 3,
            8: 1,
            9: 3,
            10: 2,
            11: 3,
            12: 2,
            13: 2,
            14: 3,
            15: 3,
            16: 1,
            17: 2,
            18: 1,
            19: 2,
            20: 1,
        }
        self.assertEqual(result, expected)

    def test_init_intesity(self):
        result = self.canalyser.init_intensity(
            ["low", "medium", "high", "very high"],
            [0, 0.7, 1.2, 2.0],
            [Fore.BLUE, Fore.YELLOW, Fore.RED, Fore.MAGENTA],
        )
        expected = [
            Intensity("low", 0, Fore.BLUE),
            Intensity("medium", 0.7, Fore.YELLOW),
            Intensity("high", 1.2, Fore.RED),
            Intensity("very high", 2.0, Fore.MAGENTA),
        ]
        self.assertEqual(result, expected)

        with self.assertRaises(DifferentListSizeError):
            self.canalyser.init_intensity(
                ["low", "medium", "high", "very high"],
                [0, 0.7, 1.2],
                [Fore.BLUE, Fore.YELLOW],
            )

        with self.assertRaises(ConstantsNotAscendingError):
            self.canalyser.init_intensity(
                ["low", "medium", "high", "very high"],
                [0, 1, 0.5, 2.0],
                [Fore.BLUE, Fore.YELLOW, Fore.RED, Fore.MAGENTA],
            )

        with self.assertRaises(ConstantsNotUniqueError):
            self.canalyser.init_intensity(
                ["low", "medium", "high", "very high"],
                [0, 1, 1, 2],
                [Fore.BLUE, Fore.YELLOW, Fore.RED, Fore.MAGENTA],
            )

    def test_calculate_moving_average(self):
        time_frequency = {}
        for i in range(20):
            time_frequency[i] = i + 1

        self.canalyser.get_frequency()
        self.canalyser.window = 4
        result = self.canalyser.calculate_moving_average()
        expected = {
            0: 0.0,
            1: 0.5,
            2: 1.6666666666666667,
            3: 1.75,
            4: 2.25,
            5: 3.25,
            6: 3.75,
            7: 4.0,
            8: 3.75,
            9: 3.25,
            10: 2.25,
            11: 2.25,
            12: 2.5,
            13: 2.25,
            14: 2.5,
            15: 2.5,
            16: 2.25,
            17: 2.25,
            18: 1.75,
            19: 1.5,
            20: 1.5,
        }

        self.assertEqual(result, expected)

        with self.assertRaises(ValueError):
            ChatAnalyser([], log_path=None, window=1, default_context_path=None)

    def test_create_highlight_annotation(self):
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        result = self.canalyser.create_highlight_annotation()
        expected = [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            0,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
            -1,
        ]
        self.assertEqual(result, expected)

    def test_detect_highlight_times(self):
        self.canalyser = ChatAnalyser(generate_random_chat(100, 101, 4), log_path=None, window=5, default_context_path=None)
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        highlights = self.canalyser.detect_highlight_times()

        result = [
            "[{}] ({}->{}): {}".format(h.stream_id, h.time, h.duration, h.fdelta)
            for h in highlights
        ]
        expected = [
            '[undefined] (1->20): 1.7000000000000006',
            '[undefined] (38->10): 0.5050000000000003', 
            '[undefined] (54->5): 0.17000000000000082',
            '[undefined] (60->8): 0.5350000000000001'
        ]

        self.assertEqual(result, expected)

    def test_correct_highlights(self):
        self.canalyser = ChatAnalyser(generate_random_chat(100, 101, 10), log_path=None, window=5, default_context_path=None)
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.detect_highlight_times()

        # manually modified a highlight to be too short
        self.canalyser.highlights[1].duration = 3
        initial_len = len(self.canalyser.highlights)

        corrected_highlights = self.canalyser.correct_highlights()

        result = [
            "[{}] ({}->{}): {}".format(h.stream_id, h.time, h.duration, h.fdelta)
            for h in corrected_highlights
        ]

        expected = [
            '[undefined] (31->3): 0.30000000000000426',
            '[undefined] (38->12): 1.045',
            '[undefined] (73->7): 0.5450000000000017'
        ]

        self.assertEqual(len(self.canalyser.highlights), len(corrected_highlights))
        self.assertNotEqual(initial_len, len(corrected_highlights))
        self.assertEqual(result, expected)

    def test_set_highlight_intensities(self):
        self.canalyser = ChatAnalyser(generate_random_chat(100, 101, 10), log_path=None, window=5, default_context_path=None)
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.detect_highlight_times()
        self.canalyser.correct_highlights()

        self.canalyser.init_intensity(
            ["low", "medium", "high", "very high"],
            [0, 0.4, 1.1, 1.7],
            [Fore.BLUE, Fore.YELLOW, Fore.RED, Fore.MAGENTA],
        )

        hl_int = self.canalyser.set_highlight_intensities()
        result = [h.colorless_str for h in hl_int]

        expected = [
            '[0:00:31] :  (No messages, medium intensity, 0.300 diff, 5s duration)',
            '[0:00:38] :  (No messages, high intensity, 1.045 diff, 12s duration)',
            '[0:01:13] :  (No messages, medium intensity, 0.545 diff, 7s duration)'
        ]

        self.assertEqual(result, expected)

    def test_get_highlight_messages(self):
        self.canalyser = ChatAnalyser(generate_random_chat(100, 101, 10), log_path=None, window=5, default_context_path=None)
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.detect_highlight_times()
        self.canalyser.correct_highlights()
        self.canalyser.init_intensity()
        self.canalyser.set_highlight_intensities()

        highlights = self.canalyser.get_highlight_messages()

        result = [hl.colorless_str for hl in highlights]

        expected = [
            '[0:00:31] :  (58 messages, medium intensity, 0.300 diff, 5s duration)',
            '[0:00:38] :  (114 messages, very high intensity, 1.045 diff, 12s duration)',
            '[0:01:13] :  (75 messages, high intensity, 0.545 diff, 7s duration)'
        ]

        self.assertEqual(result, expected)

    def test_get_highlight_keywords(self):
        self.canalyser = ChatAnalyser(generate_random_chat(100, 101, 10), log_path=None, window=5, default_context_path=None)
        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.detect_highlight_times()
        self.canalyser.correct_highlights()
        self.canalyser.init_intensity()
        self.canalyser.set_highlight_intensities()
        self.canalyser.get_highlight_messages()

        highlights = self.canalyser.get_highlight_keywords()

        result = [hl.keywords for hl in highlights]
        expected = [
            ['msg32', 'msg34', 'msg35', 'msg33'],
            ['msg40', 'msg42', 'msg49', 'msg45'],
            ['msg77', 'msg79', 'msg76', 'msg74']
        ]
        
        self.assertEqual(result, expected)

    def test_guess_context(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            window=5,
            default_context_path=SAMPLE_CONTEXT_PATH,
        )

        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.detect_highlight_times()
        self.canalyser.correct_highlights()
        self.canalyser.init_intensity()
        self.canalyser.set_highlight_intensities()
        self.canalyser.get_highlight_messages()
        self.canalyser.get_highlight_keywords()

        expected = [
            ["msg6", "msg13", "msg4", "msg19"],
            ["msg32", "msg34", "msg35", "msg33"],
            ["msg40", "msg42", "msg49", "msg45"],
            ["msg77", "msg79", "msg76", "msg74"],
        ]
        self.canalyser.get_contexts()
        highlights = self.canalyser.guess_context()
        result = [hl.contexts for hl in highlights]
        expected = [
            {"reaction3", "reaction1"},
            {"reaction2"},
            {"reaction3"},
        ]
        expected2 = [
            {"reaction1", "reaction3"},
            {"reaction2"},
            {"reaction3"},
        ]
        try:
            self.assertEqual(result, expected)
        except:
            self.assertEqual(result, expected2)

    def test_get_highlights(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            window=5,
            default_context_path=SAMPLE_CONTEXT_PATH,
        )

        self.canalyser.get_frequency()
        self.canalyser.calculate_moving_average()
        self.canalyser.smoothen_mov_avg()
        self.canalyser.create_highlight_annotation()
        self.canalyser.get_contexts()
        highlights = self.canalyser.get_highlights()
        result = [hl.colorless_str for hl in highlights]

        expected = [
            '[0:00:31] reaction1: msg32 (58 messages, medium intensity, 0.300 diff, 5s duration)',
            '[0:00:38] reaction2: msg40 (114 messages, very high intensity, 1.045 diff, 12s duration)',
            '[0:01:13] reaction3: msg77 (75 messages, high intensity, 0.545 diff, 7s duration)'
        ]
        self.assertEqual(result, expected)

    def test_analyse(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            window=5,
            default_context_path=SAMPLE_CONTEXT_PATH,
        )
        
        self.canalyser.analyse()
        result = [hl.colorless_str for hl in self.canalyser.highlights]
        expected = [
            "[0:00:31] reaction1: msg32 (58 messages, medium intensity, 0.300 diff, 5s duration)",
            "[0:00:38] reaction2: msg40 (114 messages, very high intensity, 1.045 diff, 12s duration)",
            "[0:01:13] reaction3: msg77 (75 messages, high intensity, 0.545 diff, 7s duration)",
        ]
        self.assertEqual(result, expected)

    def test_context_source_structure(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            default_context_path=None,
        )

        # RESET
        self.canalyser.source.reset()
        self.assertEqual(self.canalyser.source.paths, [])


        # ADD
        ## basic
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.add(ANOTHER_PATH)
        self.assertEqual(
            self.canalyser.source.paths,
            [SAMPLE_CONTEXT_PATH, ANOTHER_PATH]
        )
        self.canalyser.source.reset()

        ## should raise PathAlreadyExistsException if path already exists
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        with self.assertRaises(PathAlreadyExistsException):
            self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.reset()
        
        ## should raise PathAlreadyExistsException if path is None or empty
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        with self.assertRaises(ValueError):
            self.canalyser.source.add(None)
            self.canalyser.source.add("")
        self.canalyser.source.reset()


        # REMOVE
        ## by path
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.remove(byPath=SAMPLE_CONTEXT_PATH)
        self.assertEqual(self.canalyser.source.paths, [])
        self.canalyser.source.reset()

        ## by index
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.remove(byIndex=0)
        self.assertEqual(self.canalyser.source.paths, [])
        self.canalyser.source.reset()

        ## remove last with index
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.remove(byIndex=-1)
        self.assertEqual(self.canalyser.source.paths, [])
        self.canalyser.source.reset()

        # should raise value error when both parameters are used
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        with self.assertRaises(ValueError):
            self.canalyser.source.remove(
                byPath=SAMPLE_CONTEXT_PATH,
                byIndex=0
            )
        self.canalyser.source.reset()

        # should raise value error when none of the parameters are used
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        with self.assertRaises(ValueError):
            self.canalyser.source.remove()
        self.canalyser.source.reset()


        # UPDATE
        ## basic
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.source.update(SAMPLE_CONTEXT_PATH, ANOTHER_PATH)
        self.assertEqual(self.canalyser.source.paths, [ANOTHER_PATH])
        self.canalyser.source.reset()


    def test_read_contexts_from_sources(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            default_context_path=SAMPLE_CONTEXT_PATH,
        )

        # basic
        self.canalyser.read_contexts_from_sources()
        self.assertTrue(self.canalyser.contexts, sample_contexts)

        # should reset contexts each time
        self.canalyser.read_contexts_from_sources()
        self.canalyser.read_contexts_from_sources()
        self.assertTrue(self.canalyser.contexts, sample_contexts)

    def test__check_contexts(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            default_context_path=None,
        )

        # should not do anything
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.read_contexts_from_sources()
        self.canalyser._check_contexts()
        self.assertEqual(self.canalyser.contexts, sample_contexts)

        # should raise exception if duplicate context is found
        self.canalyser.source.add(ANOTHER_PATH)
        self.canalyser.read_contexts_from_sources()
        with self.assertRaises(DuplicateContextException):
            self.canalyser._check_contexts()

        # should merge duplicate contexts on `autofix = True`
        # instead of throwing error
        self.canalyser._check_contexts(autofix=True)
        
        self.assertEqual(
            self.canalyser.contexts,
            [{'reaction_to': 'reaction1', 'triggers': [{'phrase': 'msg6', 'is_exact': False}, {'phrase': 'msg34', 'is_exact': False}, {'phrase': 'msg2', 'is_exact': False}, {'phrase': 'msg32', 'is_exact': False}, {'phrase': 'msg6', 'is_exact': False}, {'phrase': 'msg34', 'is_exact': False}, {'phrase': 'msg2', 'is_exact': False}, {'phrase': 'msg32', 'is_exact': False}]}, {'reaction_to': 'reaction2', 'triggers': [{'phrase': 'msg42msg40', 'is_exact': True}, {'phrase': 'msg49', 'is_exact': True}, {'phrase': 'msg40', 'is_exact': True}, {'phrase': 'msg42msg40', 'is_exact': True}, {'phrase': 'msg49', 'is_exact': True}, {'phrase': 'msg40', 'is_exact': True}]}, {'reaction_to': 'reaction3', 'triggers': [{'phrase': 'msg74', 'is_exact': False}, {'phrase': 'msg35', 'is_exact': False}, {'phrase': 'msg77', 'is_exact': False}, {'phrase': 'msg74', 'is_exact': False}, {'phrase': 'msg35', 'is_exact': False}, {'phrase': 'msg77', 'is_exact': False}]}]
        )

    def test_parse_contexts(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            default_context_path=None,
        )
        self.canalyser.source.add(SAMPLE_CONTEXT_PATH)
        self.canalyser.read_contexts_from_sources()
        contexts = self.canalyser.parse_contexts()
        self.assertEqual(type(contexts[0]), Context)
        self.assertEqual(type(contexts[0].triggers[0]), Trigger)


        self.canalyser.source.update(SAMPLE_CONTEXT_PATH, PARTIAL_CORRUPT_PATH)
        self.canalyser.read_contexts_from_sources()

        # should raise exception when parsing corrupt data
        with self.assertRaises(Exception):
            self.canalyser.parse_contexts()

        # should skip corrupt data if autofix is enabled
        contexts = self.canalyser.parse_contexts(autofix=True)
        self.assertEqual(contexts[0].reaction_to, partially_corrupt_contexts[1]["reaction_to"])

        self.canalyser.source.update(PARTIAL_CORRUPT_PATH, FULL_CORRUPT_PATH)
        self.canalyser.read_contexts_from_sources()
        # should raise ContextsAllCorruptException if autofix is true and all the data is corrupt
        with self.assertRaises(ContextsAllCorruptException):
            self.canalyser.parse_contexts(autofix=True)
        
        # should raise another Exception depending on the source file content
        # if autofix is false and all the data is corrupt
        with self.assertRaises(Exception):
            self.canalyser.parse_contexts(autofix=False)

    def test_get_contexts(self):
        self.canalyser = ChatAnalyser(
            generate_random_chat(100, 101, 10),
            log_path=None,
            default_context_path=SAMPLE_CONTEXT_PATH,
        )

        # should return all parsed sample contexts
        self.canalyser.get_contexts()
        self.assertEqual(
            [context.reaction_to for context in self.canalyser.contexts],
            [context['reaction_to'] for context in sample_contexts],
        )


        self.canalyser.source.reset()
        self.canalyser.source.add(PARTIAL_CORRUPT_PATH)
        
        # should throw error if anything is wrong when autofix is disabled
        with self.assertRaises(KeyError):
            self.canalyser.get_contexts(autofix=False)
        
        # should return some parsed contexts when autofix is enabled
        self.canalyser.get_contexts(autofix=True)
        self.assertEqual(
            [context.reaction_to for context in self.canalyser.contexts],
            [context['reaction_to'] for context in partially_corrupt_contexts[1:]],
        )


        self.canalyser.source.reset()
        self.canalyser.source.add(FULL_CORRUPT_PATH)
        
        # should return nothing if all the data is corrupt when autofix is enabled
        self.canalyser.get_contexts(autofix=True)
        self.assertTrue(len(self.canalyser.contexts)==0)
        
        # should throw error if all the data is corrupt when autofix is disabled
        with self.assertRaises(KeyError):
            self.canalyser.get_contexts(autofix=False)
        
if __name__ == "__main__":
    unittest.main()



percentage = lambda current, out_of: round(int(current * 100 / out_of))


def normalize(text) -> str:
    """Normalizes string by removing punctuations, trimming and lowering.
    If a string only consists of punctuations, the text is trimmed only."""

    punctuation = r"""!"#$%&'()*+,-./:;<=>?@[\]^_`{|}~"""  # maybe add ？！ー～…
    if all(ch in punctuation for ch in text):
        return text.strip()
    return text.lower().strip().translate(str.maketrans("", "", punctuation))



from dataclasses import dataclass, field
import datetime
from typing import Optional
import webbrowser
import platform
from colorama import init, Fore, Style
from colorama.ansi import AnsiFore
from enum import IntEnum
#from .chatanalyser import DEFAULT_CONTEXT_SOURCE_PATH  # circular import
from .exceptions import PathAlreadyExistsException

init()

class ImageResolution(IntEnum):
    MEDIUM = 0
    HIGH = 1
    STANDARD = 2
    MAXIMUM = 3

class Browser:
    CHROME = "chrome"
    EDGE = "edge"
    FIREFOX = "firefox"
    OPERA = "opera"

class OSName:
    WINDOWS = "Windows"
    LINUX = "Linux"
    MAC = "Darwin"
    

@dataclass
class Icon:
    id: str  # title
    url: str
    width: int = 0
    height: int = 0

    def __repr__(self):
        if self.width and self.height:
            return f"{self.id} ({self.width}x{self.height}): {self.url}"
        return f"{self.id}: {self.url}"

@dataclass
class Emote:
    id: str
    name: str
    is_custom_emoji: bool
    images: list[Icon] = field(default_factory=list)

    def __hash__(self):
        return hash(self.id+self.name)

    def __repr__(self):
        return f"{self.id}: {self.name} ({len(self.images)} images)"

@dataclass
class Author:
    id: str
    name: str
    is_member: bool = False
    membership_info: str = ""
    images: list[Icon] = field(default_factory=list)

    def colorless_str(self):
        if self.is_member:
            return f"{self.name}: {self.id} [{self.membership_info}]"
        return f"{self.name}: {self.id}"

    def __repr__(self):
        color = Fore.GREEN if self.is_member else Fore.YELLOW
        if self.is_member:
            return f"{color+self.name+Style.RESET_ALL}: {self.id} [{self.membership_info}]"
        return f"{color+self.name+Style.RESET_ALL}: {self.id}"

    def __hash__(self):
        return hash(self.id)

@dataclass
class SuperchatColor:
    background: str
    header: str

    def __repr__(self):
        return f"{self.header}/{self.background}"


@dataclass
class Money:
    amount: str
    currency: str
    currency_symbol: str
    text: str

    def __repr__(self):
        return f"{self.text} ({self.currency})"

@dataclass
class ChatItem:
    id: str
    time: int
    author: Author
    text: str

    @property
    def time_in_hms(self):
        return datetime.timedelta(seconds=int(self.time))
    
    def __hash__(self):
        return hash(self.id)

@dataclass
class Message(ChatItem):
    emotes: list[Emote] = field(default_factory=list)

    @property
    def colorless_str(self):
        return f"[{self.time_in_hms}] {self.author.name}: {self.text}"

    def __repr__(self):
        return f"[{self.time_in_hms}] {Fore.YELLOW+self.author.name+Style.RESET_ALL}: {self.text}"

@dataclass
class Superchat(ChatItem):
    money: Money
    colors: SuperchatColor
    emotes: list[Emote] = field(default_factory=list)

    @property
    def colorless_str(self):
        return f"[{self.time_in_hms}] {self.author.name}: {self.text} ({self.money.text})"

    def __repr__(self):
        return f"[{self.time_in_hms}] {Fore.RED+self.author.name+Style.RESET_ALL}: {self.text} ({self.money.text})"

@dataclass
class Membership(ChatItem):
    welcome_text: str
    emotes: list[Emote] = field(default_factory=list)

    @property
    def colorless_str(self):
        return f"[{self.time_in_hms}] {self.author.name} has joined membership. {str(self.welcome_text)}"

    def __repr__(self):
        return f"[{self.time_in_hms}] {Fore.GREEN+self.author.name+Style.RESET_ALL} has joined membership. {str(self.welcome_text)}"

@dataclass
class Sticker(Superchat):
    sticker_images: list[Icon] = field(default_factory=list)

    @property
    def colorless_str(self):
        return f"[{self.time_in_hms}] {self.author.name} sent a Sticker ({self.money.text})"

    def __repr__(self):
        return f"[{self.time_in_hms}] {Fore.RED+self.author.name+Style.RESET_ALL} sent a Sticker ({self.money.text})"

@dataclass
class Intensity:
    level: str
    constant: int
    color: AnsiFore

    def __repr__(self):
        return f"{self.color+self.level+Style.RESET_ALL}: {self.constant}"

    @property
    def colored_level(self):
        return self.color + self.level + Style.RESET_ALL


@dataclass
class Highlight:
    stream_id: str
    time: int
    duration: int
    intensity: Intensity = None
    fdelta: float = 0.0  # is the frequency difference from start to finish
    messages: list[Message] = field(default_factory=list)
    keywords: list[str] = field(default_factory=list)
    kw_emotes: list[Emote] = field(default_factory=list)
    contexts: set[str] = field(default_factory=set)

    def __repr__(self):
        return "[{0}] {1}: {2} ({3} messages, {4} intensity, {5:.3f} diff, {6}s duration)".format(
            self.time_in_hms,
            Fore.LIGHTRED_EX + "/".join(self.contexts) + Style.RESET_ALL,
            ", ".join(self.keywords),
            len(self.messages) if self.messages else "No",
            self.intensity.colored_level,
            self.fdelta,
            self.duration,
        )

    @property
    def colorless_str(self):
        return "[{0}] {1}: {2} ({3} messages, {4} intensity, {5:.3f} diff, {6}s duration)".format(
            self.time_in_hms,
            "/".join(self.contexts),
            ", ".join(self.keywords),
            len(self.messages) if self.messages else "No",
            self.intensity.level,
            self.fdelta,
            self.duration,
        )

    @property
    def url(self):
        return f"https://youtu.be/{self.stream_id}?t={self.time}"

    def open_in_browser(self, browser:Browser=Browser.CHROME, browser_path=None):
        """Open highlight in browser

        Args:
            browser (str): Browser name to get default path.

            browser_path (str, optional): Path to executable
                browser file. Overrides browser argument.
                Defaults to None.
        """

        if browser_path:
            webbrowser.get(browser_path).open(self.url)
            return

        # TODO modify default path for other OS's
        if platform.system == OSName.WINDOWS:
            if browser == Browser.CHROME:
                browser_path = (
                    r"C:/Program Files/Google/Chrome/Application/chrome.exe %s"
                )
            elif browser == Browser.EDGE:
                browser_path = (
                    r"C:/Program Files (x86)/Microsoft/Edge/Application/msedge.exe %s"
                )
            elif browser == Browser.FIREFOX:
                browser_path = r"C:/Program Files/Mozilla Firefox/FireFox.exe %s"
            elif browser == Browser.OPERA:
                browser_path = r"C:/Program Files/Opera/Launcher.exe %s"
        elif platform.system == OSName.LINUX:
            pass
        elif platform.system == OSName.MAC:
            pass
        else:
            pass

        webbrowser.get(browser_path).open(self.url)

    def to_dict(self) -> dict:
        return {
            "stream_id": self.stream_id,
            "time": self.time,
            "duration": self.duration,
            "intensity": self.intensity,
            "fdelta": self.fdelta,
            "messages": self.messages,
            "keywords": self.keywords,
            "kw_emotes": self.kw_emotes,
            "contexts": self.contexts,
        }

    @property
    def time_in_hms(self):
        return datetime.timedelta(seconds=int(self.time))


@dataclass
class ContextSourceManager():
    """Dataclass to manage source paths for context files
    """
    #paths: list = field(default_factory=lambda:[DEFAULT_CONTEXT_SOURCE_PATH])
    paths: list = field(default_factory=list)
    def __repr__(self):
        return "[%s]" % ", ".join(self.paths)

    def add(self, full_path:str):
        if not full_path:
            raise ValueError("Should provide a full path")
        if full_path not in self.paths:
            self.paths.append(full_path)
            # TODO check if valid path
            return
        raise PathAlreadyExistsException(f"Warning: '{full_path}' already exists in paths")
    def remove(self, byPath:Optional[str]=None, byIndex:Optional[int]=None):
        if byPath is not None and byIndex is not None:
            raise ValueError("Can't use both byPath and byIndex parameters")
        if byPath is not None:
            self.paths.remove(byPath)
        elif byIndex is not None:
            del self.paths[byIndex]
        else:
            raise ValueError("Should provide one of the parameters")
    def update(self, old_path:str, new_path:str):
        self.remove(byPath=old_path)
        self.add(new_path)
    def reset(self):
        self.paths = []


@dataclass
class Trigger():
    phrase: str
    is_exact: bool

    def __repr__(self):
        return f"{self.phrase} ({'exact' if self.is_exact else 'inexact'})"


@dataclass
class Context():
    reaction_to: str
    triggers: list[Trigger] = field(default_factory=list)

    def __hash__(self):
        return hash(self.reaction_to)

    def __repr__(self):
        return f"{self.reaction_to}: {', '.join([str(trigger) for trigger in self.triggers])}"



import os
import json
from collections import Counter
from typing import Optional
import string

from colorama import Fore
from matplotlib import collections, pyplot as plt, font_manager as fm, rcParams
import numpy as np

from .loggersetup import create_logger
from . import utils
from .structures import (
    Emote,
    Intensity,
    Highlight,
    ContextSourceManager,
    Context,
    Trigger
)
from .exceptions import (
    DifferentListSizeError,
    ConstantsNotAscendingError,
    ConstantsNotUniqueError,
    DuplicateContextException,
    UnexpectedException,
    ContextsAllCorruptException,
    PathAlreadyExistsException
)
from .keyphrase_finder import KeyphraseFinder

DEFAULT_FONT_PATH = os.path.join(
    os.path.dirname(os.path.realpath(__file__)), "..", "fonts", "NotoSansCJKjp-Bold.ttf"
)


class ChatAnalyser:
    """A class to analyse live chat messages

    Args:
        refined_messages (list[Message]): List of messages of the stream refined by
            the DataRefiner class.

        log_path(str): Path to log folder. Set to None to log without writing to a file.

        stream_id (str, optional): Stream id of the chat. Defaults to 'undefined'.

        min_duration (int): Minimum highlight duration (in seconds) to detect. Defaults to 5

        window (int, optional):  Time interval to calculate moving averages. Defaults to 30.

        threshold_constant(int, optional): The value that divides average highlight duration.
            Set higher to get shorter highlights. Defaults to 3.

        keyword_limit(int, optional): Keyword amount to retrieve. Defaults to 4.

        keyword_filters(list, optional): Keywords to filter. Defaults to [].

        default_context_path(str, optional): Path to the default context file. Defaults to
            DEFAULT_CONTEXT_SOURCE_PATH. Set to None to disable.

        verbose (bool, optional): Make the output verbose. Defaults to False.

        stop_words_path (bool, optional): Default stop word file (.txt) path to exclude in
            keyphrase collocations. Defaults to None.
    """

    def __init__(
        self,
        refined_messages,
        log_path,
        default_context_path,
        stream_id="undefined",
        min_duration=5,
        window=30,
        threshold_constant=3,
        keyword_limit=4,
        keyword_filters=[],
        verbose=False,
        stop_words_path = None
    ):
        self.messages = refined_messages
        self.stream_id = stream_id
        self.min_duration = min_duration
        self.window = window
        self.threshold_constant = threshold_constant
        self.keyword_limit = keyword_limit
        self.keyword_filters = keyword_filters
        self.default_context_path = default_context_path
        self.verbose = verbose
        self.stop_words_path = stop_words_path
        self.logger = create_logger(__file__, log_path)

        if not self.window > 1:
            self.logger.error("Interval must be bigger than one")
            raise ValueError("Interval must be bigger than one")

        self.frequency = {}
        self.intensity_list = []
        self.fre_mov_avg = {}
        self.exp_mov_avg = []
        self.highlight_annotation = []
        self.highlights = []

        self.contexts = list[Context]
        self.source = ContextSourceManager() # important
        if self.default_context_path:
            self.source.add(self.default_context_path)

    def read_contexts_from_sources(self) -> None:
        """Reads contexts from all sources and merges them into a list"""

        self.logger.info("Reading contexts")

        self.contexts = []
        for path in self.source.paths:
            with open(path, 'r', encoding='utf-8') as file:
                data = list(json.load(file))
                self.logger.debug(f"Read {len(data)} data from {path}")
                self.contexts.extend(data)

    def _check_contexts(self, autofix:bool=False) -> None:
        """Checks context compitability. Tries to autofix collisions if possible."""

        self.logger.info("Checking contexts")
        self.logger.debug(f"{autofix=}")
        
        # TODO check triggers too (right now it only checks reactions)
        seen_tuples = set()
        new_contexts = []
        for context in self.contexts:

            # Handle key error
            try:
                context["reaction_to"]
            except KeyError:
                if autofix:
                    continue # skip data
                else:
                    err_msg = f"Key 'reaction_to' not found in: {context}"
                    self.logger.error(err_msg)
                    raise KeyError(err_msg)

            # Handle duplicate data error
            reaction_tuple = tuple({'reaction_to', context.get('reaction_to')})
            if reaction_tuple not in seen_tuples:
                seen_tuples.add(reaction_tuple)
                new_contexts.append(context)
            else: 
                if autofix:
                    self.logger.warning(f"Merging duplicate context: {context['reaction_to']}")
                    i = next((i for i, nc in enumerate(new_contexts) if nc['reaction_to'] == context['reaction_to']), None)
                    if i == None:
                        self.logger.critical(f"Unexpected error: i={i}")
                        self.logger.critical(f"{new_contexts=}")
                        self.logger.critical(f"{context=}")
                        raise UnexpectedException("You should not be seeing this.")
                    new_contexts[i].get('triggers').extend(context.get('triggers'))
                else:
                    err_msg = f"Duplicate context, reaction to '{context.get('reaction_to')}' already exists"
                    self.logger.error(err_msg)
                    raise DuplicateContextException(
                        err_msg,
                        encounters=[] # TODO point at the source file/files
                    )
        self.contexts = new_contexts

    def parse_contexts(self, autofix:bool=False) -> list[Context]:
        """Parses the contexts from dictionary to dataclass format"""

        self.logger.info("Parsing contexts")

        parsed_contexts = []
        for context in self.contexts:
            try:
                parsed_contexts.append(
                    Context(
                        reaction_to=context['reaction_to'],
                        triggers=[
                            Trigger(
                                phrase=trigger['phrase'],
                                is_exact=trigger['is_exact']
                            ) for trigger in context['triggers']
                ]))
            except Exception as e:
                if autofix:
                    self.logger.warning(f"Skipping corrupt context data: {context}")
                    self.logger.warning(f"{e.__class__.__name__}: {e}")
                else:
                    self.logger.critical(f"Error parsing context: {context}")
                    self.logger.critical(f"{e.__class__.__name__}: {e}")
                    raise e

        if autofix and len(parsed_contexts) == 0 and len(self.contexts) != 0:
            self.logger.critical("Could not autofix, all contexts are corrupted")
            raise ContextsAllCorruptException(
                "Could not parse contexts, turn off autofix or check the log files to see the errors"
            )
        return parsed_contexts

    def get_contexts(self, autofix:bool=False) -> list[Context]:
        """Wrapper function to get the contexts"""
        self.read_contexts_from_sources()
        self._check_contexts(autofix=autofix)
        self.contexts = self.parse_contexts(autofix=autofix)
        return self.contexts

    def get_frequency(self) -> dict:
        """Creates frequency table of messages"""

        self.logger.info("Calculating frequency")

        # create frequency table
        message_frequency = {}
        for i, message in enumerate(self.messages):
            if self.verbose:
                print(
                    f"Calculating frequency...{round(utils.percentage(i, len(self.messages))/2)}%",
                    end="\r",
                )
            if message.time in message_frequency:
                message_frequency[message.time] += 1
            else:
                message_frequency[message.time] = 1

        # fill the blank seconds
        for sec in range(self.messages[-1].time):
            if self.verbose:
                print(
                    f"Calculating frequency...{round(utils.percentage(sec, self.messages[-1].time)/2)+50}%",
                    end="\r",
                )
            if not sec in message_frequency.keys():
                message_frequency[sec] = 0

        # sort
        self.frequency = {}
        for key in sorted(message_frequency.keys()):
            self.frequency[key] = message_frequency[key]
        if self.verbose:
            print(f"Calculating frequency... done")

        return self.frequency

    def init_intensity(self, levels=[], constants=[], colors=[]) -> list[Intensity]:
        """Returns list of intensity which will be used for
            measuring how tense was the highlight. Leave empty
            for default values.

        Args:
            levels (list): Names of the intensity levels.
            constants (list): Constants of the intensity levels.
            colors (list[AnsiFore]): Display colors of the intensity levels from colorama.

        Raises:
            DifferentListSizeError
            ConstantsNotAscendingError
            ConstantsNotUniqueError

        Returns:
            list[Intensity]: Intensity list
        """

        self.logger.info("Initializing intensity")
        self.logger.debug(f"{levels=}")
        self.logger.debug(f"{constants=}")
        self.logger.debug(f"{colors=}")

        levels = levels or ["medium", "high", "very high", "ultra high"]
        constants = constants or [0, 0.7, 1.4, 2.1]
        colors = colors or [Fore.YELLOW, Fore.BLUE, Fore.RED, Fore.MAGENTA]

        if len(levels) != len(constants) != len(colors):
            self.logger.error("All lists should be the same size")
            raise DifferentListSizeError("All lists should be the same size")

        if constants != sorted(constants):
            self.logger.error("Constants should be in ascending order")
            raise ConstantsNotAscendingError("Constants should be in ascending order")

        if len(set(constants)) != len(constants):
            self.logger.error("All constants should be unique")
            raise ConstantsNotUniqueError("All constants should be unique")

        self.intensity_list = [
            Intensity(levels[i], constants[i], colors[i]) for i in range(len(levels))
        ]
        return self.intensity_list

    def calculate_moving_average(self) -> dict:
        """Returns moving average of a table"""

        self.logger.info("Calculating moving average")
        self.fre_mov_avg = {}
        stack = []  # holds frequency of the last {window} seconds
        for time, value in self.frequency.items():
            if self.verbose:
                print(
                    f"Calculating moving average...{utils.percentage(time, len(self.frequency))}%",
                    end="\r",
                )
            if len(stack) == self.window:
                stack.pop(0)
            stack.append(value)
            self.fre_mov_avg[time] = sum(stack) / len(stack)

        if self.verbose:
            print(f"Calculating moving average... done")
        return self.fre_mov_avg

    def _smoothen(self, dict, w=40) -> list[np.ndarray]:
        return list(np.convolve(list(dict.values()), np.ones(w) / w, mode="same"))

    def smoothen_mov_avg(self) -> list[np.ndarray]:
        self.exp_mov_avg = self._smoothen(self.fre_mov_avg)
        return self.exp_mov_avg

    def create_highlight_annotation(self) -> list[int]:
        """Creates highlight annotation from moving average.
        Values are either -1, 0 or 1 where 1 means it's increasing."""

        self.logger.info("Creating highlight annotation")
        self.highlight_annotation = []
        for i in range(len(self.exp_mov_avg)):
            if i == len(self.exp_mov_avg) - 1:
                break
            if self.exp_mov_avg[i] < self.exp_mov_avg[i + 1]:
                self.highlight_annotation.append(1)
            elif self.exp_mov_avg[i] > self.exp_mov_avg[i + 1]:
                self.highlight_annotation.append(-1)
            else:
                self.highlight_annotation.append(0)

        for state, notation in {
            "increasing": 1,
            "decreasing": -1,
            "constant": 0,
        }.items():
            count = self.highlight_annotation.count(notation)
            self.logger.debug(f"Total {state} duration: {count}")

        return self.highlight_annotation

    def line_colors(self) -> list[str]:
        """Sets plot colors according to highlight annotation"""

        self.logger.info("Setting line colors")

        colors = []
        for x in self.highlight_annotation:
            if x == 1:
                colors.append("g")
            elif x == -1:
                colors.append("r")
            else:
                colors.append("gray")
        return colors

    def detect_highlight_times(self) -> list[Highlight]:
        """Detects highlight times and durations according to highlight annotation and
            smoothened moving average.  Also sets frequency delta, which is the change
            of frequency within highlight duration.

        Args:
            highlight_annotation (list[int]): Highlight annotation returned from
                create_highlight_annotation method.
            exp_mov_avg (list): Smoothened values of moving average of message frequency.
            min_duration (int): Minimum highlight duration to detect.
        Returns:
            list[Highlight]: List of highlight times
        """

        # TODO improve algorithm
        self.logger.info("Detecting highlight times")

        self.highlights = []
        start_time = 0
        initial_frequency = 0
        for current_time in range(len(self.highlight_annotation)):
            if self.verbose:
                print(
                    f"Detecting highlight timestamps... {utils.percentage(current_time, len(self.highlight_annotation))}%",
                    end="\r",
                )
            if not start_time and self.highlight_annotation[current_time] == 1:
                start_time = current_time
                initial_frequency = self.exp_mov_avg[current_time]

            if start_time and self.highlight_annotation[current_time] != 1:
                duration = current_time - start_time
                if duration < self.min_duration:
                    self.logger.debug(
                        f"Highlight @{start_time} was not added, duration was {duration}"
                    )
                    start_time = 0
                    continue
                delta = self.exp_mov_avg[current_time] - initial_frequency
                if delta < 0:
                    self.logger.debug(
                        f"Highlight @{start_time} was not added, delta was {delta}"
                    )
                    start_time = 0
                    continue
                self.highlights.append(
                    Highlight(self.stream_id, start_time, duration, fdelta=delta)
                )
                self.logger.debug(
                    f"Highlight found: from {start_time} to {current_time} ({duration}s)"
                )
                start_time = 0
        if self.verbose:
            print("Detecting highlight timestamps... done")
        return self.highlights

    def correct_highlights(self) -> list[Highlight]:
        """Corrects highlights by removing highlights that are too short or filtered"""

        # TODO consider fdelta too
        self.logger.info("Correcting highlights")

        if not self.highlights:
            return []

        if len(self.highlights) != 1:
            self.highlights.pop(0)

        avg_highlight_duration = sum([hl.duration for hl in self.highlights]) / len(
            self.highlights
        )
        for i, highlight in enumerate(self.highlights):
            if self.verbose:
                print(
                    f"Correcting highlights... {utils.percentage(i, len(self.highlights))}",
                    end="\r",
                )
            if highlight.duration <= avg_highlight_duration / self.threshold_constant:
                self.highlights.remove(highlight)
                if self.verbose:
                    self.logger.debug(
                        f"Removed highlight at {highlight.time}, duration was too short ({highlight.duration}s)"
                    )
        if self.verbose:
            print("Correcting highlights... done")
        return self.highlights

    def set_highlight_intensities(self) -> list[Highlight]:
        """Sets highlight intensities based on frequency delta"""

        if not self.highlights:
            return []

        self.logger.info("Setting highlight intensities")
        avg_value = sum([hl.fdelta for hl in self.highlights]) / len(self.highlights)
        for i, highlight in enumerate(self.highlights):
            if self.verbose:
                print(
                    f"Setting highlight intensities... {utils.percentage(i, len(self.highlights))}%",
                    end="\r",
                )
            for intensity in self.intensity_list:
                if highlight.fdelta > avg_value * intensity.constant:
                    highlight.intensity = intensity
            self.logger.debug(
                f"[{highlight.time}] => {highlight.intensity.level} ({highlight.fdelta})"
            )
        if self.verbose:
            print("Setting highlight intensities... done")
        return self.highlights

    def get_highlight_messages(self) -> list[Highlight]:
        """Gets messages typed during highlights"""

        self.logger.info("Getting highlight messages")

        if not self.highlights:
            return []

        hl_idx = 0
        for message in self.messages:
            if self.verbose:
                print(
                    f"Getting highlight messages... {utils.percentage(hl_idx, len(self.highlights))}%",
                    end="\r",
                )

            if (
                self.highlights[hl_idx].time
                < message.time
                < self.highlights[hl_idx].duration + self.highlights[hl_idx].time
            ):
                self.highlights[hl_idx].messages.append(message)

            if (
                message.time
                >= self.highlights[hl_idx].duration + self.highlights[hl_idx].time
            ):
                hl_idx += 1

            if hl_idx == len(self.highlights):
                break

        if self.verbose:
            print("Getting highlight messages... done")
        return self.highlights

    @staticmethod
    def get_keyword_emotes(highlight:Highlight) -> set[Emote]:
        emotes = set()
        for msg in highlight.messages:
            for emote in msg.emotes:
                if emote:
                    emotes.add(emote)
        return emotes

    def get_highlight_keywords(self) -> list[Highlight]:
        """
        Adds most frequently used words to the highlight list."""
        print()
        print("Warning: `get_highlight_keywords` is deprecated, use `get_highlight_keyphrases` instead")
        print()

        self.logger.info("Getting keywords (old function)")

        if not self.highlights:
            return []

        for i, highlight in enumerate(self.highlights):

            if self.verbose:
                print(
                    f"Getting highlight keywords... {utils.percentage(i, len(self.highlights))}%",
                    end="\r",
                )

            words = []
            if highlight.messages:
                for message in highlight.messages:
                    for word in list(set(message.text.split(" "))):
                        normalized = utils.normalize(word)
                        if normalized:
                            words.append(normalized)

            for filter in self.keyword_filters:
                try:
                    while True:
                        words.remove(filter)
                except:
                    pass
            for filter in self.keyword_filters:
                try:
                    while True:
                        words.remove(filter.upper())
                except:
                    pass

            for _word, count in Counter(words).most_common(self.keyword_limit):
                if _word and count > 1:
                    highlight.keywords.append(_word)

            if not highlight.keywords:
                self.logger.debug(
                    f"No keyword found @{highlight.time}, removing highlight"
                )
                self.highlights.remove(highlight)
            else:
                self.logger.debug(
                    f"Keywords found @{highlight.time}: {highlight.keywords}"
                )

        if self.verbose:
            print("Getting highlight keywords... done")
        return self.highlights

    def get_highlight_keyphrases(self) -> list[Highlight]:
        """Adds most frequently used phrases to the highlight list."""

        self.logger.info("Getting keyphrases")

        if not self.highlights:
            return []

        finder = KeyphraseFinder(
            chat = [],
            fix_phrases = [
                ("let 's", "let's"),
                ("ca n't", "can't"),
                ("do n't", "don't"),
                ("he 's", "he's"),
                ("they 're", "they're"),
                ("it 's", "it's"),
                ("it ’ s", "it’s"),
                ("i 'm", "i'm"),
                ("you 're", "you're"),
            ],
            punctuation_list = list(string.punctuation) + ["！","？"],
            stop_words_path = self.stop_words_path
        )

        for i, highlight in enumerate(self.highlights):
            if self.verbose:
                print(
                    f"Getting highlight keyphrases... {utils.percentage(i, len(self.highlights))}%",
                    end="\r",
                )

            if not highlight.messages:
                return self.highlights

            finder.chat = highlight.messages
            ### TODO
            # pass arguments
            # implement keyphrase class
            keywords = [r[0] for r in [tup for tup in finder.ngram_keyphrase_analysis(
                max_keyphrase_amount=self.keyword_limit,
                min_keyphrase_amount=self.keyword_limit, # to achieve non-dynamic keyphrase amount 
            )]]

            if keywords:
                highlight.keywords = keywords
                self.logger.debug(
                    f"Keyphrases found @{highlight.time}: {highlight.keywords}"
                )
                highlight.kw_emotes = list(ChatAnalyser.get_keyword_emotes(highlight))
            else:
                self.highlights.remove(highlight)
                self.logger.debug(
                    f"No keyphrase found @{highlight.time}, removing highlight"
                )

        if self.verbose:
            print("Getting highlight keyphrases... done")
        return self.highlights

    def guess_context(self) -> list[str]:
        """Guesses context by looking up the keywords for each highlight."""

        self.logger.info("Guessing context")
        if not self.highlights:
            return

        for i, highlight in enumerate(self.highlights):
            if self.verbose:
                print(
                    f"Guessing contexts... {utils.percentage(i, len(self.highlights))}%",
                    end="\r",
                )
            for keyword in highlight.keywords:
                for context in self.contexts:
                    for trigger in context.triggers:
                        if  (trigger.is_exact and trigger.phrase == keyword.lower()) or \
                            (not trigger.is_exact and trigger.phrase in keyword.lower()):
                                highlight.contexts.add(context.reaction_to)
            if not highlight.contexts:
                highlight.contexts = set(["None"])
            self.logger.debug(
                f"Guessed contexts @{highlight.time}: {highlight.contexts} from keywords"
            )

        if self.verbose:
            print(f"Guessing contexts... done")
        return self.highlights

    def get_highlights(self, autofix_context_collision:bool=False) -> list[Highlight]:
        """Returns a filled highlight list."""

        self.detect_highlight_times()
        self.correct_highlights()
        self.init_intensity()
        self.set_highlight_intensities()
        self.get_highlight_messages()
        self.get_highlight_keyphrases()
        self.get_contexts(autofix=autofix_context_collision)
        self.guess_context()
        return self.highlights

    def draw_graph(self, title=None) -> plt:
        """Draws a basic graph of the analysed data including:
        - Message frequency
        - Moving average of message frequency
        - Highlights
        """

        # TODO make a better looking graph

        self.logger.info("Drawing graph")
        if self.verbose:
            print(f"Drawing graph...", end="\r")

        # TODO background img
        # img = mpimg.imread(f"{self.config['path-to']['thumbnails']}\\{self.stream_id}.jpg")

        fig, ax = plt.subplots(2, constrained_layout=True)

        # fm.fontManager.addfont(DEFAULT_FONT_PATH)
        # rcParams['font.family'] = [fm.get_font(DEFAULT_FONT_PATH).family_name]
        fprop = fm.FontProperties(fname=DEFAULT_FONT_PATH)
        fig.suptitle(title, fontproperties=fprop, fontsize=16)

        xAxis = list(self.frequency)
        yAxis = self.exp_mov_avg
        lines = [
            ((x0, y0), (x1, y1))
            for x0, y0, x1, y1 in zip(xAxis[:-1], yAxis[:-1], xAxis[1:], yAxis[1:])
        ]
        colors = self.line_colors()
        colored_lines = collections.LineCollection(
            lines, colors=colors, linewidths=(2,)
        )
        ax[0].add_collection(colored_lines)
        ax[0].autoscale_view()
        ax[0].set_title("Highlights")

        yAxis = list(self.frequency.values())
        ax[1].bar(xAxis, yAxis)
        yAxis = list(self.fre_mov_avg.values())
        ax[1].plot(xAxis, yAxis, "m--")
        ax[1].set_title("Message frequency")

        if self.verbose:
            print(f"Drawing graph... done")

        self.fig = plt
        return plt

    def analyse(self, levels=None, constants=None, colors=None, autofix_context_collision:bool=False):
        self.get_frequency()
        self.calculate_moving_average()
        self.smoothen_mov_avg()
        self.create_highlight_annotation()
        self.detect_highlight_times()
        self.correct_highlights()
        self.init_intensity(levels, constants, colors)
        self.set_highlight_intensities()
        self.get_highlight_messages()
        self.get_highlight_keyphrases()
        self.get_contexts(autofix=autofix_context_collision)
        self.guess_context()



from os import stat
from typing import Tuple
from urllib import request, parse
import json

from yaml.events import DocumentStartEvent

from chat_downloader import ChatDownloader, errors
from chat_downloader.sites.youtube import YouTubeChatDownloader 
import isodate

from .structures import ImageResolution
from .loggersetup import create_logger
from .utils import percentage
from .exceptions import StreamIsLiveOrUpcomingError


class DataCollector:
    """A class that fetches required data to analyse the stream."""

    def __init__(self, id, log_path, msglimit=None, verbose=False, yt_api_key=None) -> None:
        self.id = id
        self.logger = create_logger(__file__, log_path, sid=id)
        if self._is_live_or_upcoming:
            raise StreamIsLiveOrUpcomingError("Stream needs to be archived first to get its messages")
        self._check_chat_replay()

        self.msglimit = msglimit
        self.verbose = verbose
        self.yt_api_key = yt_api_key

        self.iscomplete = False
        self.metadata = {}

    def _check_chat_replay(self):
        try:
            ChatDownloader().get_chat("https://www.youtube.com/watch?v=" + self.id, max_messages=1)
        except errors.NoChatReplay:
            self.logger.error(f"Chat replay is not available: https://www.youtube.com/watch?v={self.id}")
            raise errors.NoChatReplay("Chat replay is not available")

    def collect_metadata(self) -> dict:
        """Collects metadata of the YouTube stream"""

        self.logger.info("Collecting metadata")
        if self.verbose:
            print(f"Collecting metadata...", end="\r")

        data = self._get_oembed_respone()
        data['duration'] = self._get_video_duration()

        self.metadata = data
        if self.verbose:
            print(f"Collecting metadata... done")
        return data

    def _get_video_duration(self) -> int:
        if not self.yt_api_key:
            try:
                return int(YouTubeChatDownloader().get_video_data(self.id).get("duration"))
            except Exception as e:
                self.logger.error(f"Couldn't get video duration, returning -1 instead. ({e.__class__.__name__}: {e})")
                return -1
        return self._parse_duration(
            json.load(
                request.urlopen(
                    f"https://www.googleapis.com/youtube/v3/videos?id={self.id}&key={self.yt_api_key}&part=contentDetails"
                )
            )["items"][0]["contentDetails"]["duration"]
        )   

    @staticmethod
    def _parse_duration(yt_duration_response) -> int:
        return int(isodate.parse_duration(yt_duration_response).total_seconds())

    def _get_oembed_respone(self) -> dict:
        params = {
            "format": "json",
            "url": "https://www.youtube.com/watch?v=%s" % self.id,
        }
        url = "https://www.youtube.com/oembed"
        query_string = parse.urlencode(params)
        url = url + "?" + query_string
        try:
            with request.urlopen(url) as response:
                response_text = response.read()
                data = json.loads(response_text.decode())
        except request.URLError as e:
            self.logger.critical("Couldn't get oembed info. Returning empty metadata instead.")
            return {
                "title":"None",
                "author_name":"None",
                "author_url":"None",
                "type":"None",
                "height":-1,
                "width":-1,
                "version":"None",
                "provider_name":"None",
                "provider_url":"None",
                "thumbnail_height":-1,
                "thumbnail_width":-1,
                "thumbnail_url":"None",
                "html":"None"
            }
        except Exception as e:
            self.logger.error(e)
            raise e
        return data

    def _enforce_time_consistency(self, messages) -> Tuple[list, int]:
        """Quick fix for a bug where ChatDownloader module sometimes
        returns messages that are written LONG AFTER (~5 hours) the
        stream ends by removing the last message.
        """
        inconsistent_data_amount = 0
        if "duration" in self.metadata.keys():
            if not self.metadata["duration"]:
                self.logger.warning("Can't check time consistency as duration is not determined yet")
                return messages, 0
            while messages[-1]["time_in_seconds"] > self.metadata["duration"]:
                self.logger.warning(f"Deleted message as its time was exceeding the video length: {messages[-1]['time_in_seconds']} ({self.metadata['duration']})")
                inconsistent_data_amount+=1
                del messages[-1]
        return messages, inconsistent_data_amount

    @property
    def _is_live_or_upcoming(self) -> bool:
        """Returns if the stream is live or upcoming""" # lol
        return YouTubeChatDownloader().get_video_data(self.id).get("status") != 'past'

    def fetch_raw_messages(self) -> list[dict]:
        """Fetches live chat messages"""

        self.logger.info("Fetching messages")
        raw_messages = []
        yt_url = "https://www.youtube.com/watch?v=" + self.id
        corrupted_data_amount = 0
        try:
            for counter, raw_message in enumerate(
                ChatDownloader().get_chat(yt_url, start_time=0, message_groups=['messages', 'superchat']), start=1
            ):
                if self.verbose:
                    print(
                        f"Fetching raw messages... {str(percentage(counter, self.msglimit))+'%' if self.msglimit else counter}",
                        end="\r",
                    )
                try:
                    raw_messages.append(self._reformat_message(raw_message))
                except KeyError as e:
                    self.logger.warning(f"Corrupt message data skipped: {raw_message}")
                    corrupted_data_amount += 1
                    continue
                if self.msglimit and counter == self.msglimit:
                    break
        except Exception as e:
            print(e)
            self.logger.critical(
                f"Could not fetch messages: {e.__class__.__name__}:{e}"
            )
            raise e

        self.iscomplete = not bool(self.msglimit)
        raw_messages, inconsistent_data_amount = self._enforce_time_consistency(raw_messages)

        if self.verbose:
            print(f"Fetching raw messages... done")

        self.logger.info(
            f"{len(raw_messages)-corrupted_data_amount} messages fetched ({corrupted_data_amount} corrupted, {inconsistent_data_amount} inconsistent)"
        )

        return raw_messages

    @staticmethod
    def _reformat_message(message) -> dict:
        """Reformats messages returned from ChatDownloader."""
        reformatted_message = {
            "message_id": message["message_id"],
            "message_type": message["message_type"],
            "message": message["message"],
            "time_in_seconds": message["time_in_seconds"],
            "author": {
                "name": message["author"]["name"],
                "id": message["author"]["id"],
                "images": message["author"]["images"],
            }
        }
        if "emotes" in message.keys():
            reformatted_message["emotes"] = [{
                "id": emote["id"],
                "name": emote["name"],
                "images": emote["images"],
                "is_custom_emoji": emote["is_custom_emoji"]
            } for emote in message["emotes"]]
        if "badges" in message.get("author").keys():
            reformatted_message["author"]["badges"] = message["author"]["badges"]

        # Important: To get memberships and superchats, follow https://github.com/xenova/chat-downloader/pull/126/commits/14637d3f5a75fc97f480aa4f3732dde0134221ee
        if message.get("message_type") == "paid_message":
            reformatted_message["money"] = {
                "amount": message["money"]["amount"],
                "currency": message["money"]["currency"],
                "currency_symbol": message["money"]["currency_symbol"],
                "text": message["money"]["text"],
            }
            reformatted_message["colors"] = {
                "body_background_colour": message["body_background_colour"],
                "header_background_colour": message["header_background_colour"],
            }
        elif message.get("message_type") == "membership_item":
            reformatted_message["welcome_text"] = str(message["header_secondary_text"]),
        elif message.get("message_type") == "paid_sticker":
            reformatted_message["money"] = {
                "amount": message["money"]["amount"],
                "currency": message["money"]["currency"],
                "currency_symbol": message["money"]["currency_symbol"],
                "text": message["money"]["text"],
            }
            reformatted_message["colors"] = {
                "body_background_colour": message["background_colour"],
                "header_background_colour": None,
            }
            reformatted_message["sticker_images"] = message["sticker_images"]
        return reformatted_message
            
    def fetch_missing_messages(self, start_time, current_amount, target_amount=None) -> list[dict]:
        """Returns missing messages.

        Args:
            start_time (float): Starting time to fetch messages.
            current_amount (int): Current message amount.
            target_amount (int): Target message amount. Defaults to None,
                which returns all remaining messages.

        Returns:
            list[dict]: List of missing messages.
        """

        self.logger.info("Fetching missing messages")
        self.logger.debug(f"{start_time=}")
        self.logger.debug(f"{current_amount=}")
        self.logger.debug(f"{target_amount=}")

        yt_url = "https://www.youtube.com/watch?v=" + self.id
        corrupted_data_amount = 0
        raw_messages = []
        limit = current_amount + target_amount - 1 if target_amount else None

        for counter, raw_message in enumerate(
            ChatDownloader().get_chat(yt_url, start_time=start_time),
            start=current_amount,
        ):
            if self.verbose:
                print(
                    f"Fetching missing messages... {str(percentage(counter, limit))+'%' if target_amount else counter}",
                    end="\r",
                )
            try:
                raw_messages.append(self._reformat_message(raw_message))
            except KeyError:
                self.logger.warning(f"Corrupt message data skipped: {raw_message}")
                corrupted_data_amount += 1
                continue
            if target_amount and counter == limit:
                break

        if not self.iscomplete:
            self.iscomplete = not bool(target_amount)

        if self.verbose:
            print(f"Fetching missing messages... done")

        self.logger.info(
            f"{len(raw_messages)-corrupted_data_amount} messages fetched ({corrupted_data_amount} corrupted)"
        )

        return raw_messages

    def get_thumbnail_url(self, res_lvl:ImageResolution=ImageResolution.STANDARD) -> str:
        """Gets URL of the thumbnail image.

        Args:
            res_lv (ImageResolution, optional): Resolution level of the thumbnail. Defaults to `ImageResolution.STANDARD`.

        Returns:
            str: URL of the thumbnail image.

        """
        self.logger.info("Getting thumbnail url")
        self.logger.debug(f"{res_lvl=}")

        res_lvls = ["mqdefault", "hqdefault", "sddefault", "maxresdefault"]

        if not 0 <= res_lvl < 4:
            self.logger.warning("res_lvl was out of range, set it to 2")
            res_lvl = 2

        return f"https://i.ytimg.com/vi/{self.id}/{res_lvls[res_lvl]}.jpg"



import unittest
import warnings

from modules.datarefiner import DataRefiner
from modules.structures import Icon, Message, Author


class TestDataRefiner(unittest.TestCase):
    def test_refine_raw_messages(self):
        # w/out limit
        result_messages = self.refiner.refine_raw_messages(
            self.sample_raw_messages, msglimit=None
        )
        self.assertEqual(result_messages, self.expected_messages)

        # w/ limit
        result_messages = self.refiner.refine_raw_messages(
            self.sample_raw_messages, msglimit=3
        )
        self.assertEqual(result_messages, self.expected_messages[:3])

        result_messages = self.refiner.refine_raw_messages(
            [{"corrupt_data": "test"}],
        )
        self.assertEqual(result_messages, [])

    def test_get_authors(self):
        result_authors = self.refiner.get_authors()
        self.assertEqual(result_authors, [])

        self.refiner.refine_raw_messages(self.sample_raw_messages)
        result_authors = self.refiner.get_authors()
        self.assertEqual(result_authors, self.expected_authors)

    def setUp(self):
        warnings.simplefilter("ignore", category=ResourceWarning)
        self.refiner = DataRefiner(log_path=None)
        self.refiner.logger.disabled = True

        self.sample_raw_messages, self.expected_messages, self.expected_authors = [], [], []
        for i in range(6):
            self.sample_raw_messages.append({
                    "message_id": str(i),
                    "message_type": "text_message",
                    "message": "msg" + str(i),
                    "time_in_seconds": i,
                    "author": {
                        "id": str(i),
                        "name": "name" + str(i),
                        "images": [{
                            "id": str(i),
                            "url": "url" + str(i), 
                        }],
                    },
                }
            )

            author = Author(
                id=str(i),
                name="name"+str(i),
                images={
                    "profile": [Icon(id=str(i), url="url"+str(i))],
                    "membership": []
                },
            )
            self.expected_authors.append(author)
            self.expected_messages.append(
                Message(
                    id=str(i),
                    text="msg"+str(i),
                    time=i,
                    author=author
                )
            )
    def tearDown(self):
        del self.refiner


if __name__ == "__main__":
    unittest.main()



import json
import shutil
import unittest
import warnings
import os
import importlib

import sys
if __package__ is None:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from modules.structures import Author, Message
from streamanalyser import streamanalyser as sa 

class TestStreamAnalyser(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", category=ResourceWarning)

    def tearDown(self):
        # clear caches of used id's
        for id in ["testid", "um196SMIoR8"]:
            with sa.StreamAnalyser(id, 0, disable_logs=True) as analyser:
                analyser.clear_cache()

    def test_cache_messages(self):
        with sa.StreamAnalyser("testid", 1, disable_logs=True) as analyser:
            analyser._cache_messages([{}])
            self.assertTrue(
                os.path.isfile(
                    os.path.join(
                        analyser.filehandler.sid_path,
                        analyser.filehandler.message_fname + ".gz",
                    )
                )
            )

    def test_cache_metadata(self):
        with sa.StreamAnalyser("testid", 1, disable_logs=True) as analyser:
            analyser._cache_metadata({})

            self.assertTrue(
                os.path.isfile(
                    os.path.join(
                        analyser.filehandler.sid_path,
                        analyser.filehandler.metadata_fname,
                    )
                )
            )

    def test_collect_read_data(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.collect_data()
            analyser.read_data()
            self.assertTrue(
                analyser._raw_messages[0]["author"]["id"], "UCX07ffYvacTkgo89MjNpweg"
            )
            self.assertTrue(
                analyser.metadata["title"],
                "【APEX】英語…頑張ります！！\u3000w/" + " Roboco,Amelia【常闇トワ/ホロライブ】",
            )

    @unittest.skip("Fix later. Refactor sample messages.")
    def test_refine_data(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser._raw_messages = sample_raw_messages
            analyser.refine_data()
            self.assertEqual(analyser.messages[0].text, "kon...")
            self.assertEqual(analyser.authors[0].id, "UCX07ffYvacTkgo89MjNpweg")

    @unittest.skip("Fix later. Refactor sample messages.")
    def test_analyse_data(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.messages = analyser.refiner.refine_raw_messages(
                sample_raw_messages
            )
            analyser.metadata = {"title": "test"}
            analyser.analyse_data()

            self.assertEqual(
                analyser.highlights[0].colorless_str,
                "[0:00:01] greeting: こんやっぴートワ様いえーい, こん, こん…, "
                + "kon (24 messages, high intensity, 1.042 diff, 19s duration)",
            )

    @unittest.skipIf(
        importlib.util.find_spec("not_a_module") is None,
        "WordCloud module is not installed",
    )
    def test_generate_wordcloud(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.messages = analyser.refiner.refine_raw_messages(
                sample_raw_messages
            )
            self.assertTrue(analyser.generate_wordcloud(scale=0.1).to_image())

    def test_find_messages(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.messages = [Message(
                id=raw_message["message_id"],
                text=raw_message["message"],
                time=round(raw_message["time_in_seconds"]),
                author=Author(
                    id=raw_message["author"]["id"],
                    name=raw_message["author"]["name"],
                ),
            ) for raw_message in sample_raw_messages]

            self.assertEqual(
                analyser.find_messages("こんやっぴ~")[0].colorless_str,
                "[0:00:39] Yuzuchu: こんやっぴ~",
            )

            # ignore_case should produce different results
            self.assertNotEqual(
                analyser.find_messages("kon...", ignore_case=False),
                analyser.find_messages("kon...", ignore_case=True),
            )

            # It should return empty list when exact is set
            # even though there's "kon" in other messages
            self.assertEqual(analyser.find_messages("kon", exact=True), [])

    def test_find_user_messages(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.messages = [Message(
                id=raw_message["message_id"],
                text=raw_message["message"],
                time=round(raw_message["time_in_seconds"]),
                author=Author(
                    id=raw_message["author"]["id"],
                    name=raw_message["author"]["name"],
                ),
            ) for raw_message in sample_raw_messages]

            self.assertEqual(len(analyser.find_user_messages(username="Yuzuchu")), 2)

            self.assertEqual(
                len(analyser.find_user_messages(id="UCBLOc9HL4kIvp36bMqu7pxg")), 1
            )

    def test_most_used_phrase(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.messages = [Message(
                id=raw_message["message_id"],
                text=raw_message["message"],
                time=round(raw_message["time_in_seconds"]),
                author=Author(
                    id=raw_message["author"]["id"],
                    name=raw_message["author"]["name"],
                ),
            ) for raw_message in sample_raw_messages]

            self.assertEqual(analyser.most_used_phrase(), ("こんやっぴートワ様いえーい", 35))

            # with not normalize
            self.assertEqual(
                analyser.most_used_phrase(normalize=False), ("こん:_やっぴー::_トワ様いえーい:", 35)
            )

            # with exclude
            self.assertEqual(
                analyser.most_used_phrase(
                    exclude=["こんやっぴートワ様いえーい"],
                ),
                ("konyappi", 11),
            )

            # with exclude and not normalize
            self.assertEqual(
                analyser.most_used_phrase(
                    exclude=["こん:_やっぴー::_トワ様いえーい:"], normalize=False
                ),
                ("こん...", 6),
            )

    def test_analyse(self):
        src_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),  "custom_context.json"
        )
        with open(src_path, "w", encoding="utf-8") as file:
            file.write(json.dumps([
                {
                    "reaction_to": "greeting",
                    "triggers": [
                        {
                            "phrase": "こん :_やっぴー::_トワ様いえーい:",
                            "is_exact": True
                        },
                        {
                            "phrase": "konyappi",
                            "is_exact": False
                        },
                    ]
                }], indent=4
            ))
        # BUG: fix keyword limit error (set keyword limit to 10 to reproduce)
        with sa.StreamAnalyser("um196SMIoR8", 100, disable_logs=True, keyword_limit=10, default_context_path=None) as analyser:
            analyser.context_source.add(src_path)
            analyser.analyse()
            self.assertEqual(analyser.highlights[0].contexts, {"greeting"})

            os.remove(src_path)

    def test_check_integrity(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.collect_data()

            # intentionally delete metadata file
            analyser.filehandler.delete_file(
                os.path.join(
                    analyser.filehandler.sid_path, analyser.filehandler.metadata_fname
                )
            )

            # create unnecessary dummy file
            with open(analyser.filehandler.sid_path + "\\dummy.txt", "w"):
                pass

            missing_files, unnecessary_files = analyser.filehandler.check_integrity()

            # should detect metadata as missing file
            self.assertEqual(missing_files, [analyser.filehandler.metadata_fname])

            # should detect dummy.txt as unnecessary file
            self.assertEqual(unnecessary_files, ["dummy.txt"])

            # intentionally decompress message file
            analyser.filehandler._decompress_file(
                os.path.join(
                    analyser.filehandler.sid_path, analyser.filehandler.message_fname
                )
            )

            # try again with autofix
            missing_files, unnecessary_files = analyser.filehandler.check_integrity(
                autofix=True
            )

            # still should detect metadata as missing file,
            # but decompressed file should be compressed
            self.assertEqual(missing_files, [analyser.filehandler.metadata_fname])
            # should delete dummy file
            self.assertEqual(unnecessary_files, [])

    def test_enforce_integrity(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            analyser.collect_data()

            # intentionally delete cache files
            analyser.clear_cache(delete_root_folder=False)

            # create unnecessary dummy file
            with open(analyser.filehandler.sid_path + "\\dummy.txt", "w+") as f:
                pass

            # call the function
            analyser.enforce_integrity()

            # check integrity now
            missing_files, unnecessary_files = analyser.filehandler.check_integrity()

            self.assertEqual(missing_files, [])
            self.assertEqual(unnecessary_files, [])

    @unittest.skipIf(
        importlib.util.find_spec("not_a_module") is None,
        "WordCloud module is not installed",
    )
    def test_export(self):
        with sa.StreamAnalyser("um196SMIoR8", 1, disable_logs=True) as analyser:
            msg_path = os.path.join(
                analyser.filehandler.sid_path,
                analyser.filehandler.message_fname + ".gz",
            )
            mdata_pth = os.path.join(
                analyser.filehandler.sid_path,
                analyser.filehandler.metadata_fname,
            )

            analyser._raw_messages = sample_raw_messages
            analyser.refine_data()
            analyser.highlights = [
                sa.structures.Highlight(
                    "testid",
                    0,
                    10,
                    sa.structures.Intensity("lvl", 2, sa.structures.AnsiFore.GREEN),
                    5,
                    analyser.messages,
                    ["kword"],
                    ["contxt"],
                )
            ]

            with open(msg_path, "w"):
                pass
            with open(mdata_pth, "w"):
                pass

            analyser.export_data(folder_name="test235234")
            test_export_folder = os.path.join(
                analyser.filehandler.export_path, "test235234"
            )
            fnames = analyser.filehandler.get_filenames(test_export_folder)
            cheklist = ["highlights", "messages", "metadata", "thumbnail", "wordcloud"]
            for item in cheklist:
                self.assertTrue(item in fnames)

            shutil.rmtree(test_export_folder)

    def test_fetch_missing_messages(self):
        with sa.StreamAnalyser("um196SMIoR8", 2, disable_logs=True) as analyser:
            analyser.collect_data()
            analyser.read_data()
            analyser.refine_data()

            # increase message limit (inital was 2)
            analyser.msglimit = 4

            # call the function
            analyser.fetch_missing_messages()

            # the missing 2 should be fetched and refined
            self.assertEqual(len(analyser.messages), 4)

            # set is-complete to True manually
            # (it's set to true by itself when msg_limit is None in practice)
            analyser.update_metadata({"is-complete": True})

            # increase message limit again
            analyser.msglimit = 8

            # call the function
            analyser.fetch_missing_messages()

            # the messages should NOT be fetched as is-complete is `True`
            self.assertEqual(len(analyser.messages), 4)

    @unittest.skip(
        "This test is always skipped for now as it clears existing cache. Gonna fix it later"
    )
    def test_cache_options(self):
        with sa.StreamAnalyser(
            "testid", 0, cache_limit=10, disable_logs=False
        ) as analyser:
            # create dummy files in the cache folder
            # and pass cache limit intentionally
            msg_path = os.path.join(
                analyser.filehandler.cache_path,
                "testid",
                analyser.filehandler.message_fname + ".gz",
            )
            with open(msg_path, "w"):
                pass
            for i in range(analyser.cache_limit):
                analyser.filehandler.create_dir_if_not_exists(
                    os.path.join(
                        analyser.filehandler.cache_path,
                        "testid" + str(i),
                    )
                )
            msg_path = os.path.join(
                analyser.filehandler.cache_path,
                "testid" + str(i),
                analyser.filehandler.message_fname + ".gz",
            )
            with open(msg_path, "w"):
                pass

        with sa.StreamAnalyser(
            "testid",
            0,
            disable_logs=False,
            cache_limit=10,
            cache_deletion_algorithm="mru",
        ) as analyser:
            # most recently created folder must have been deleted
            self.assertFalse(
                os.path.exists(
                    os.path.join(
                        analyser.filehandler.cache_path,
                        "testid" + str(analyser.cache_limit - 1),
                    )
                )
            )

        with sa.StreamAnalyser(
            "testid",
            0,
            disable_logs=False,
            cache_limit=9,
            cache_deletion_algorithm="fifo",
        ) as analyser:
            # least recently created folder must have been deleted
            self.assertFalse(
                os.path.exists(os.path.join(analyser.filehandler.cache_path, "testid0"))
            )

        with sa.StreamAnalyser(
            "testid",
            0,
            disable_logs=False,
            cache_limit=8,
            cache_deletion_algorithm="lru",
        ) as analyser:
            # least recently created folder must have been deleted
            self.assertFalse(
                os.path.exists(os.path.join(analyser.filehandler.cache_path, "testid1"))
            )

sample_raw_messages = [
    {
        "author": {"id": "UCX07ffYvacTkgo89MjNpweg", "name": "RathalosRE"},
        "message_type": "text_message",
        "message": "kon...",
        "message_id": "CjkKGkNMSHU0ZmpxNl9FQ0ZjTWZmUW9kWlJzSXFBEhtDSlRCd0szbTZfRUNGY2ZGeEFvZHVVUUEwZzA%3D",
        "time_in_seconds": 0,
    },
    {
        "author": {"id": "UClaAJUb11yYCjY5PdhZmfsg", "name": "rubin"},
        "message_type": "text_message",
        "message": ":_トワ文字:iting:_トワ様いえーい:",
        "message_id": "CjkKGkNObWgtX2pxNl9FQ0ZZVVo1d29keGl3SlBREhtDTTZqcnVqcTZfRUNGY2VBWXdZZC1HSU5jdzA%3D",
        "time_in_seconds": 0,
    },
    {
        "author": {"id": "UCApQ-9DiTSIeqTZWXMXzR8g", "name": "Lucas Barreto"},
        "message_type": "text_message",
        "message": "こん。。。",
        "message_id": "CkUKGkNLN1g0UHZxNl9FQ0ZWb1ByUVlkVFg4RHVBEidDSmlENl9mcTZfRUNGWWlKbVFvZHNOOEwxQTE2MjY1ODQzNjgwNzU%3D",
        "time_in_seconds": 2.042,
    },
    {
        "author": {"id": "UCBLOc9HL4kIvp36bMqu7pxg", "name": "Rena Moon"},
        "message_type": "text_message",
        "message": "こん",
        "message_id": "CjkKGkNPcmVtZnpxNl9FQ0ZhbkZ3UW9kem4wQ1pREhtDSlBaLUpqcDZfRUNGWVhaMVFvZFJCc0FxZzA%3D",
        "time_in_seconds": 4.126,
    },
    {
        "author": {"id": "UCK89HIgAk_WVwbl8dySd1ng", "name": "Prince Chips"},
        "message_type": "text_message",
        "message": ":_トワ文字:aiting!",
        "message_id": "CjkKGkNQZXh6ZnpxNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDT0Npel9YcTZfRUNGUVZUV0FvZHJZRUYzdzA%3D",
        "time_in_seconds": 5.334,
    },
    {
        "author": {"id": "UCyesjldaXmoLOKRoBaekf_w", "name": "Chewyboot "},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CjkKGkNPYXJsZjNxNl9FQ0ZZRVhmUW9kWHY4SU53EhtDSm1qX3VUTDZfRUNGYllUclFZZEZiNEhVQTE%3D",
        "time_in_seconds": 6.414,
    },
    {
        "author": {"id": "UCm3PZqgUsSa5t4X_KU8JLLQ", "name": "Silence voice"},
        "message_type": "text_message",
        "message": "kon...",
        "message_id": "CjkKGkNOUzM0ZjNxNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDTHZVamZmazZfRUNGZDBMdHdBZGhqc0lYUTE%3D",
        "time_in_seconds": 7.262,
    },
    {
        "author": {"id": "UCRGu8LiQgoN3rSaSGAFWHlg", "name": "kurobear"},
        "message_type": "text_message",
        "message": "こん....",
        "message_id": "CjkKGkNMRHU3ZjNxNl9FQ0ZkZ0FyUVlkdVBRRlFREhtDTlB6d3ZYcTZfRUNGYWRSaFFvZENKc0czZzA%3D",
        "time_in_seconds": 7.405,
    },
    {
        "author": {"id": "UCuPkyqDfoIHYl-sSfhdDtDw", "name": "Yuzuchu"},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CkUKGkNPaUNtZjdxNl9FQ0ZkZ0FyUVlkdVBRRlFREidDTU9vekxfcTZfRUNGZWJEd2dRZGpiY0k4dzE2MjY1ODQzNzM4MDM%3D",
        "time_in_seconds": 8.207,
    },
    {
        "author": {"id": "UCb6Kbp7NtVgeSD9a9DCz1nQ", "name": "Yumina"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CjkKGkNOQzctdjdxNl9FQ0ZSd2RyUVlkb25FQnR3EhtDSUdNM01mbzZfRUNGYlNybFFJZHg4QUhYdzE%3D",
        "time_in_seconds": 10.182,
    },
    {
        "author": {"id": "UCw6g9yBkt54jUJI6elkdZzg", "name": "Khoa Nguyễn Văn"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CkUKGkNMT0tydl9xNl9FQ0ZjTWZmUW9kWlJzSXFBEidDUDNaMGZ2cTZfRUNGWHRTaFFvZGhNQUYwQTE2MjY1ODQzNzQxNzk%3D",
        "time_in_seconds": 11.014,
    },
    {
        "author": {"id": "UCojDEDgA2-m632NaR29vNdQ", "name": "XxKawaiixRicexX"},
        "message_type": "text_message",
        "message": "Welcome to the family:_トワ様いえーい:",
        "message_id": "CjkKGkNJR1Rzdl9xNl9FQ0ZkZ0FyUVlkdVBRRlFREhtDTGZWcEpqcDZfRUNGWXlpeEFvZE56b1AxUTA%3D",
        "time_in_seconds": 11.051,
    },
    {
        "author": {"id": "UCgTvHUoeeF8z_pFqbr8e26w", "name": "mayt_ S."},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNQRC1sSURyNl9FQ0ZjX0N3Z1FkUHl3RnRREidDT2lpbXZ2cTZfRUNGYm5BT0FZZGhZQU5uUTE2MjY1ODQzNzY4ODE%3D",
        "time_in_seconds": 12.797,
    },
    {
        "author": {"id": "UCx4WCmY3Z09P9383-W49yig", "name": "如月白兎"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNPUDJ5WURyNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDTFA3czdmZjZfRUNGZWxDOVFVZHR1UUh5ZzI%3D",
        "time_in_seconds": 13.564,
    },
    {
        "author": {"id": "UCLeLH-PQA2ADFOW1AorGUZw", "name": "Hannah Althea"},
        "message_type": "text_message",
        "message": "Kon....",
        "message_id": "CjkKGkNNYTdob0hyNl9FQ0ZjX0N3Z1FkUHl3RnRREhtDTWI3OF96cTZfRUNGVU5IaFFvZFRBVUdydzA%3D",
        "time_in_seconds": 14.211,
    },
    {
        "author": {"id": "UCSTO1VvCNyHoMoEQhScEJIw", "name": "Raseru Kiiro"},
        "message_type": "text_message",
        "message": "Konyappi !",
        "message_id": "CjkKGkNMYkFrSUhyNl9FQ0ZaRU5yUVlkQ0I4UEpnEhtDS0dreDR2VDZfRUNGWW9rV0FvZFdLWUxWUTA%3D",
        "time_in_seconds": 14.32,
    },
    {
        "author": {"id": "UCp5N3jaxHlMq-Vu7OLJOJ9A", "name": "Rain_sw ⁸⁰れいん"},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CjoKGkNNMjl5NEhyNl9FQ0ZWb1ByUVlkVFg4RHVBEhxDUHJjN0lqbzZfRUNGY1hWVEFJZFVad0xKUS0x",
        "time_in_seconds": 15.766,
    },
    {
        "author": {"id": "UCX07ffYvacTkgo89MjNpweg", "name": "RathalosRE"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNNdlQxb0hyNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDSlRCd0szbTZfRUNGY2ZGeEFvZHVVUUEwZzE%3D",
        "time_in_seconds": 15.902,
    },
    {
        "author": {"id": "UCpHWyGyWROHCbaKna0AdWsg", "name": "Big On Mars"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNJR2g5NEhyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDTm1lNEtmcTZfRUNGUnVKeEFvZGRLUUU2ZzE%3D",
        "time_in_seconds": 16.389,
    },
    {
        "author": {"id": "UCtSNv6VgO9Wx7YN8yqlbnMg", "name": "miha 398"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CjoKGkNMN2JvNExyNl9FQ0ZjTWZmUW9kWlJzSXFBEhxDSmItNGVYcTZfRUNGUUx4T0FZZEhINE1CQS0y",
        "time_in_seconds": 17.228,
    },
    {
        "author": {"id": "UCZ0TEGOiLpOZOAOxA_nRmXw", "name": "きっき"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNJT0J1WUxyNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDTERHMi1ETDZfRUNGUlN4V0FvZHd0RUVhQTE%3D",
        "time_in_seconds": 17.444,
    },
    {
        "author": {"id": "UCo7o-Ib_3NMyNAU-062vN2A", "name": "Old Can"},
        "message_type": "text_message",
        "message": "KONYAPPI",
        "message_id": "CjkKGkNOTzIxSUxyNl9FQ0ZZRVhmUW9kWHY4SU53EhtDT3lialBmcTZfRUNGZGxSWUFvZHZoRU9sZzE%3D",
        "time_in_seconds": 17.965,
    },
    {
        "author": {"id": "UCNDoIL08saEOTZqshRDbOwQ", "name": "ゆきにゃん"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CjoKGkNKcmEzNExyNl9FQ0ZkZ0FyUVlkdVBRRlFREhxDT3pNLV9fcTZfRUNGYWRDaFFvZFBqd01Xdy0w",
        "time_in_seconds": 18.157,
    },
    {
        "author": {"id": "UCiuFMPV1e-fgacO5-uRIwmg", "name": "OMEGAエイキ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNQTGc1b0xyNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDTWJ1MzY3cTZfRUNGVlFGV0FvZExka09UQTE%3D",
        "time_in_seconds": 18.187,
    },
    {
        "author": {"id": "UC61Dme0yj5Y_zd7j1YmFsVQ", "name": "てってれけ"},
        "message_type": "text_message",
        "message": "こんやっぴ～",
        "message_id": "CjkKGkNQT0w2SUxyNl9FQ0ZaTTFyUVlkRTVvRnRBEhtDTDZNbE5IVTZfRUNGWWhZWUFvZEFZc0RwZzA%3D",
        "time_in_seconds": 18.315,
    },
    {
        "author": {"id": "UCyesjldaXmoLOKRoBaekf_w", "name": "Chewyboot"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNLbk1xWVByNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDSm1qX3VUTDZfRUNGYllUclFZZEZiNEhVQTI%3D",
        "time_in_seconds": 19.355,
    },
    {
        "author": {"id": "UC2GrTQsYbNO2T10jgs7qMZg", "name": "Michael Rowland"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNQZjUyb1ByNl9FQ0ZST2JnZ29kMFU4T3hnEhtDUGlTb29YWjZfRUNGWVdvbkFvZEVkZ0xUdzE%3D",
        "time_in_seconds": 20.226,
    },
    {
        "author": {"id": "UCmVWpWJQWmdneqUHIIu2lgw", "name": "JULIAN HERMES 276"},
        "message_type": "text_message",
        "message": "konyappi",
        "message_id": "CkUKGkNOS0lpb1RyNl9FQ0ZlQU5yUVlkMTJrS1FBEidDUFNmcklIcjZfRUNGWGpKY3dFZEdySUcwQTE2MjY1ODQzODQ1ODE%3D",
        "time_in_seconds": 21.06,
    },
    {
        "author": {"id": "UCZZCqNj8D-sttddThLZivUA", "name": "Jimmy Axe"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNPZVVtWVRyNl9FQ0ZZRVhmUW9kWHY4SU53EidDS2VVNm9EcjZfRUNGWXN6S2dvZHBfRURuQTE2MjY1ODQzODUxMjk%3D",
        "time_in_seconds": 21.119,
    },
    {
        "author": {"id": "UCefjOmhflYyZdhU_UHFbqOg", "name": "LetDownLad"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNOVG9ub1RyNl9FQ0ZWb1ByUVlkVFg4RHVBEidDTXZaNG9IcjZfRUNGUVFPclFZZHpIa0NlZzE2MjY1ODQzODQ2NjM%3D",
        "time_in_seconds": 21.233,
    },
    {
        "author": {"id": "UCBJSteJ5LvA8mxoRj0hb_QQ", "name": "avie"},
        "message_type": "text_message",
        "message": "konyappi",
        "message_id": "CkUKGkNManF5b1RyNl9FQ0ZaTTFyUVlkRTVvRnRBEidDS0xmNjREcjZfRUNGYWZhY3dFZHhpc0daUTE2MjY1ODQzODU2MTc%3D",
        "time_in_seconds": 21.942,
    },
    {
        "author": {"id": "UC4ocycPs14eBRTDKC-ZM7kA", "name": "Hololive World"},
        "message_type": "text_message",
        "message": "Hi",
        "message_id": "CkUKGkNKTDV5SVRyNl9FQ0ZlQU5yUVlkMTJrS1FBEidDSzJ3aW9QcjZfRUNGVlRLeEFvZEhpUURVQTE2MjY1ODQzODU3MTk%3D",
        "time_in_seconds": 21.957,
    },
    {
        "author": {"id": "UCsNn93pMaxPR8X7cFtvwZ8g", "name": "Jeje Rndnw"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNJNi15b1RyNl9FQ0ZZRVhmUW9kWHY4SU53EhtDSzJzN0x2YzZfRUNGUkxnY3dFZFV6MElSZzE%3D",
        "time_in_seconds": 22.008,
    },
    {
        "author": {"id": "UCxUSC7aLHt86aMkMUbkgI1g", "name": "Asanya"},
        "message_type": "text_message",
        "message": "こん。。。",
        "message_id": "CkUKGkNORDV5NFRyNl9FQ0ZZbUt3UW9ka1NNRXpREidDTWpLNGYzcTZfRUNGWm1YVVFvZDFKc0xIQTE2MjY1ODQzODg1NjI%3D",
        "time_in_seconds": 22.042,
    },
    {
        "author": {"id": "UCjxi9NW5768uhHAPhdjVW7g", "name": "ソリダス・スネーク"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNKbU1nSVhyNl9FQ0ZjVU5yUVlkQnlZQ3BREhtDTWlUcWJiWjZfRUNGWGhXaFFvZDZCVUwzdzE%3D",
        "time_in_seconds": 22.907,
    },
    {
        "author": {"id": "UCOlCkAoREDtM5JtpKdz8iyQ", "name": "Epuration"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNQcnl3WVhyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDT2ZyeF83bzZfRUNGVk1VZlFvZHhSVU0zZzE%3D",
        "time_in_seconds": 23.958,
    },
    {
        "author": {"id": "UCqbDXM6JwWc2raPEHW4hqCQ", "name": "カズユキ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNJLXp6SVhyNl9FQ0ZkZ0FyUVlkdVBRRlFREidDTWk5cl83cTZfRUNGYmYtT0FZZGFCMEljZzE2MjY1ODQzODk1MzQ%3D",
        "time_in_seconds": 24.031,
    },
    {
        "author": {"id": "UCQ7IlLZ2xLtH5MBZe2jqu3w", "name": "ゆっくりゆあめ"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CkUKGkNQX3ExNFhyNl9FQ0ZjX0N3Z1FkUHl3RnRREidDSkM3X3ZfcTZfRUNGUmd3WUFvZGluVUJXUTE2MjY1ODQzODc0MzQ%3D",
        "time_in_seconds": 24.231,
    },
    {
        "author": {"id": "UCwFklj_pbj0vOxkMev02I4A", "name": "Zheng Xuan Lim"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNQSzc4NFhyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDSjNyczdIYjZfRUNGVXFnNlFvZG5Da09WdzE%3D",
        "time_in_seconds": 24.736,
    },
    {
        "author": {"id": "UCzH8uSW84hwa79oH6Kigaig", "name": "ヴィラガグレン"},
        "message_type": "text_message",
        "message": "こんやっぴ～",
        "message_id": "CjkKGkNNLVZfb1hyNl9FQ0ZjTWZmUW9kWlJzSXFBEhtDTm1Kc09IVTZfRUNGWlJIaFFvZEttWUhiUTE%3D",
        "time_in_seconds": 25.038,
    },
    {
        "author": {"id": "UCINfAMONpumYpwDBFY5ekjw", "name": "しょーにゅー"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNNci1pb2JyNl9FQ0ZSd2RyUVlkb25FQnR3EhtDT2FEN1BQcDZfRUNGVUxqVEFJZHE1SU9XZzI%3D",
        "time_in_seconds": 25.128,
    },
    {
        "author": {"id": "UCqQxu2t-O2kqtPKy54Oz7Qg", "name": "JH"},
        "message_type": "text_message",
        "message": "🎶",
        "message_id": "CjkKGkNQN3NpWWJyNl9FQ0ZaTTFyUVlkRTVvRnRBEhtDSlNoNmV6cTZfRUNGUnhKVEFnZFYtOEMzQTA%3D",
        "time_in_seconds": 25.156,
    },
    {
        "author": {"id": "UC-GyPOPmWcOYUIDd_OYtXCA", "name": "Jajammy RXGD"},
        "message_type": "text_message",
        "message": "konyappi!",
        "message_id": "CkUKGkNPenBub2JyNl9FQ0ZjVU5yUVlkQnlZQ3BREidDTzJjZ29McjZfRUNGYm5MVEFJZDFuQU9OQTE2MjY1ODQzODY3MDI%3D",
        "time_in_seconds": 25.495,
    },
    {
        "author": {"id": "UCLqvMB03DmAEBl4G8qzymhQ", "name": "氷翠"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNQYmtyb2JyNl9FQ0ZWb1ByUVlkVFg4RHVBEidDT0x4NjRMcjZfRUNGVTVFOVFVZHNZQUpYZzE2MjY1ODQzODkzOTc%3D",
        "time_in_seconds": 25.788,
    },
    {
        "author": {"id": "UCatjAKjmySNRrVYvri1bIng", "name": "Nachtflut"},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CkUKGkNNRGl0NGJyNl9FQ0ZRN2F3UW9kb0dJQXFREidDT1hNc2NqcTZfRUNGUWZhVlFvZENxd0FmZzE2MjY1ODQzOTE2OTI%3D",
        "time_in_seconds": 25.993,
    },
    {
        "author": {"id": "UCFqD6L3nhesaQtRmqt_s6aA", "name": "Winks 8"},
        "message_type": "text_message",
        "message": "Kon…",
        "message_id": "CjoKGkNJS1kxb2JyNl9FQ0ZaTTFyUVlkRTVvRnRBEhxDTlRjLTRIcjZfRUNGUm5zT0FZZDZISURhdy0w",
        "time_in_seconds": 26.387,
    },
    {
        "author": {"id": "UC6iKIVMw_zuQWpvAHOyioRA", "name": "いつでも甘いものが食べたい"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNPdU0zWWJyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDSnVtcHFQYjZfRUNGUUhXVEFJZHZQVUNSQTA%3D",
        "time_in_seconds": 26.408,
    },
    {
        "author": {"id": "UCJYmnUdNODP6oDn9FDLkIvQ", "name": "Nekomusume"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNJU24yNGJyNl9FQ0ZjVU5yUVlkQnlZQ3BREidDS19PaTRQcjZfRUNGZHdWWUFvZHJERUFHUTE2MjY1ODQzOTA4NDA%3D",
        "time_in_seconds": 26.409,
    },
    {
        "author": {"id": "UC9l04tt9qHc9tw65pjPJgvQ", "name": "Alterchemist"},
        "message_type": "text_message",
        "message": ":_トワ様いえーい::_トワ様いえーい::_トワ様いえーい:",
        "message_id": "CjkKGkNMZVMzNGJyNl9FQ0ZRcUp3UW9kajI0QjBBEhtDTTZILTdUaDZfRUNGVUVFWlFvZHdKVUw3QTE%3D",
        "time_in_seconds": 26.576,
    },
    {
        "author": {"id": "UCTd2o7B6znKsv_iVAbrQSRA", "name": "upa Upa"},
        "message_type": "text_message",
        "message": "こんやっぴ～",
        "message_id": "CjkKGkNQZTc3SWJyNl9FQ0ZZVVo1d29keGl3SlBREhtDSTM1cklfUzZfRUNGYzB2WUFvZGlVa0ZwZzA%3D",
        "time_in_seconds": 26.657,
    },
    {
        "author": {"id": "UCjsVbOmWfGDv_p4Ju0lJ54w", "name": "Ksha"},
        "message_type": "text_message",
        "message": "konyappi!",
        "message_id": "CjkKGkNQamw4NGJyNl9FQ0ZjX0N3Z1FkUHl3RnRREhtDTmJaeGNfcTZfRUNGYVV6dHdBZHE0NE5YZzA%3D",
        "time_in_seconds": 26.816,
    },
    {
        "author": {"id": "UCt_Cqg78Dg3Lj70vkEytmkA", "name": "Luis de vera"},
        "message_type": "text_message",
        "message": "Konyappi",
        "message_id": "CkUKGkNKNzktSWJyNl9FQ0ZZRVhmUW9kWHY4SU53EidDSWZuMzRIcjZfRUNGVTYzV0FvZE9Qb0J6QTE2MjY1ODQzOTMxNTc%3D",
        "time_in_seconds": 26.989,
    },
    {
        "author": {"id": "UCbcED252E4O3ImCavRrcOmA", "name": "Yamato Pendragon"},
        "message_type": "text_message",
        "message": "こん…",
        "message_id": "CjkKGkNPcVprb2ZyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDTHI3NzkzazZfRUNGWmVBWXdZZGlKc0lkZzA%3D",
        "time_in_seconds": 27.322,
    },
    {
        "author": {"id": "UCnofPFAHCQRx6RCSRMocnQQ", "name": "Unlimitedpensel Ch."},
        "message_type": "text_message",
        "message": "we are always family",
        "message_id": "CkUKGkNJbVZwWWZyNl9FQ0ZjVU5yUVlkQnlZQ3BREidDTWFreFlEcjZfRUNGUUw4andvZHBiNE1RUTE2MjY1ODQzOTEzODM%3D",
        "time_in_seconds": 27.636,
    },
    {
        "author": {"id": "UCTpD5MToIioDyGd_CG-_iBA", "name": "キリシェ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjoKGkNLandwSWZyNl9FQ0ZZRVhmUW9kWHY4SU53EhxDSkhiejRMcjZfRUNGY1NpNlFvZDNMZ0s5QS0w",
        "time_in_seconds": 27.644,
    },
    {
        "author": {"id": "UCS-1flTu56hfDpOd_vHy0PQ", "name": "Ruler"},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CkUKGkNKNjRxNGZyNl9FQ0ZjX0N3Z1FkUHl3RnRREidDTG5ZeGNIcTZfRUNGV3ZRY3dFZE10d01SdzE2MjY1ODQzNTg3OTI%3D",
        "time_in_seconds": 27.772,
    },
    {
        "author": {"id": "UCIng4BBVeYpFOs1ggKjFCpw", "name": "同志ジューコフ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNOYTJ0NGZyNl9FQ0ZjX0N3Z1FkUHl3RnRREhtDSkxobGF2bjZfRUNGVXRLaXdvZFpfY09iUTA%3D",
        "time_in_seconds": 27.895,
    },
    {
        "author": {"id": "UCJ_c9jsI3eWK5UYhHkCqZ8A", "name": "しみず"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNMUHh4SWZyNl9FQ0ZSUWRmUW9kU0dBQVFBEhtDUFBZcC1YcTZfRUNGUnBMWUFvZDRLUUI2UTA%3D",
        "time_in_seconds": 28.054,
    },
    {
        "author": {"id": "UCawGDcS4AlCizJkXM8DvFSg", "name": "まな板の上のぼんじり"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjoKGkNJZWYxNGZyNl9FQ0ZZRVhmUW9kWHY4SU53EhxDT0c5Z3ZQcTZfRUNGVU1aS2dvZG9Tb0l2dy0x",
        "time_in_seconds": 28.441,
    },
    {
        "author": {"id": "UCvL9iPvI7j_ph7eV37tltSw", "name": "renchan"},
        "message_type": "text_message",
        "message": "konyappi",
        "message_id": "CkUKGkNNbmJfNGZyNl9FQ0ZZVVo1d29keGl3SlBREidDSTZfcElYcjZfRUNGVVJDblFrZF8yOEVudzE2MjY1ODQzOTMwMzk%3D",
        "time_in_seconds": 29.171,
    },
    {
        "author": {"id": "UCw6g9yBkt54jUJI6elkdZzg", "name": "Khoa Nguyễn Văn"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNMYTFnNGpyNl9FQ0ZaTTFyUVlkRTVvRnRBEidDT19vaVlUcjZfRUNGWll4V0FvZFR1b0hQdzE2MjY1ODQzOTIzOTI%3D",
        "time_in_seconds": 29.199,
    },
    {
        "author": {"id": "UC1VH0O1b9QgCOIH1X7xseyA", "name": "Davide Ciappeddu"},
        "message_type": "text_message",
        "message": "こん:_やっぴー:!!",
        "message_id": "CjkKGkNLVG1oNGpyNl9FQ0ZZbUt3UW9ka1NNRXpREhtDSmFoaklmZzZfRUNGZGhGZWdVZERvOE9EUTE%3D",
        "time_in_seconds": 29.343,
    },
    {
        "author": {"id": "UCP0xmqQQz_3zTLFjlnaltnQ", "name": "Eternal DIVE PART II"},
        "message_type": "text_message",
        "message": "KON:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNJVEprb2pyNl9FQ0ZaRU5yUVlkQ0I4UEpnEhtDTDJYOE1ITTZfRUNGVXpBb0FJZHlsQUNfUTA%3D",
        "time_in_seconds": 29.358,
    },
    {
        "author": {"id": "UCkeXQfnuNYw8L0XEt2jFYow", "name": "夕焼け"},
        "message_type": "text_message",
        "message": "こんやっぴー",
        "message_id": "CjoKGkNMZmRvWWpyNl9FQ0ZkZ0FyUVlkdVBRRlFREhxDSi1NaDRYcjZfRUNGVGZIVEFJZDFUY0hhZy0w",
        "time_in_seconds": 29.612,
    },
    {
        "author": {"id": "UCLCzX3g8v7L31mdtUj-RpiA", "name": "大チャン丸"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNMN0pySWpyNl9FQ0ZkTVE1d29kZFdVTnBREidDUHpFb1lYcjZfRUNGY05BS2dvZFNYUUp2dzE2MjY1ODQzOTMzMjc%3D",
        "time_in_seconds": 29.793,
    },
    {
        "author": {"id": "UCaNDV8TEahSX5thUaomnkeQ", "name": "CireLink"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNOWExwb2pyNl9FQ0ZiVE93UW9kbzlZTzF3EhtDUHVrdklYZDZfRUNGWXdFaVFvZC1SQUI3dzA%3D",
        "time_in_seconds": 29.809,
    },
    {
        "author": {"id": "UCfdspicLNAaO62ZiAnSJaVg", "name": "Spherical Cow in Vacuum"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNLal9xSWpyNl9FQ0ZjakZ3UW9kbkVJUFBREhtDTkNJZ1liZzZfRUNGY1Jsc2dvZFY2Y1A0dzE%3D",
        "time_in_seconds": 29.819,
    },
    {
        "author": {"id": "UCrQxhRKwzS7VQ205o2swdww", "name": "Iruka Syake Σ"},
        "message_type": "text_message",
        "message": "楽しみやー",
        "message_id": "CjoKGkNPbnBzb2pyNl9FQ0ZZVVo1d29keGl3SlBREhxDTlBDMjRMcjZfRUNGU2NOclFZZHZhWUQ3QS0w",
        "time_in_seconds": 29.932,
    },
    {
        "author": {"id": "UCzZBdhci4gqqOkyjsGclqCg", "name": "シャイニング💠ゼタ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNKYjF2b2pyNl9FQ0ZaTTFyUVlkRTVvRnRBEidDSnVjb29QcjZfRUNGYm9DdHdBZGxKZ0pMQTE2MjY1ODQzOTM1OTc%3D",
        "time_in_seconds": 30.131,
    },
    {
        "author": {"id": "UCBzODiX5d9H0hI_qbe4ywQA", "name": "wareルラ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNNNzN1NGpyNl9FQ0ZVUFh3UW9kc1pzRC13EhtDTHplcklQcjZfRUNGVVJsbUFvZEVMSUQ0dzA%3D",
        "time_in_seconds": 30.142,
    },
    {
        "author": {"id": "UCnDdoNtN-HW4XzmRvTX-axQ", "name": "Dark Rose"},
        "message_type": "text_message",
        "message": "Konyappi!:_トワ様いえーい:",
        "message_id": "CkUKGkNPN3M2NGpyNl9FQ0ZVUFh3UW9kc1pzRC13EidDS0tpcW9YcjZfRUNGVTBGMWdBZHcyQUYydzE2MjY1ODQzOTUxNTE%3D",
        "time_in_seconds": 30.63,
    },
    {
        "author": {"id": "UCTdoDKnYYBvoOWYozk7CRow", "name": "Zain"},
        "message_type": "text_message",
        "message": "こんやっぴー",
        "message_id": "CjkKGkNJV2QtNGpyNl9FQ0ZjX0N3Z1FkUHl3RnRREhtDTm0yMl9UbTZfRUNGWmJBb0FJZElwTUtjZzE%3D",
        "time_in_seconds": 31.175,
    },
    {
        "author": {"id": "UCTWtqmbyTK_JZ7_vUOGTlZQ", "name": "Lost Lenore"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNMN2lwWW5yNl9FQ0ZkTVE1d29kZFdVTnBREhtDTnZhc2VmcTZfRUNGY2dydHdBZEFkQUNRUTA%3D",
        "time_in_seconds": 31.862,
    },
    {
        "author": {"id": "UCDtd_p0SMZg__9oDL0XKxLg", "name": "白神山桜桃"},
        "message_type": "text_message",
        "message": "こんにちは",
        "message_id": "CjkKGkNOVzlyb25yNl9FQ0ZZRVhmUW9kWHY4SU53EhtDT2YzanQ3UjZfRUNGVVdPandvZGNVZ0R2dzA%3D",
        "time_in_seconds": 32.035,
    },
    {
        "author": {"id": "UCCcJWBqb36oV2y8puusQuFg", "name": "#くろまゆ"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様キラキラ:",
        "message_id": "CkUKGkNOckh1b25yNl9FQ0ZkZ0FyUVlkdVBRRlFREidDSkNXbUlmcjZfRUNGUnFfV0FvZGRFa0czZzE2MjY1ODQzOTYwNzY%3D",
        "time_in_seconds": 32.161,
    },
    {
        "author": {"id": "UCETr-d9EXiTNAWFyLtuf4-A", "name": "V Smash"},
        "message_type": "text_message",
        "message": "Towa ",
        "message_id": "CjkKGkNKM3J5NG5yNl9FQ0ZVd01zd0FkQ2hZRFpBEhtDTkt1aGZfcTZfRUNGZUU1c3dBZFg2QUQ0dzA%3D",
        "time_in_seconds": 32.498,
    },
    {
        "author": {"id": "UCuElP0-rAl3tUQw-Tvr1ZXA", "name": "烏龍"},
        "message_type": "text_message",
        "message": "こん:_やっぴー:",
        "message_id": "CjkKGkNQS1AwNG5yNl9FQ0ZjTWZmUW9kWlJzSXFBEhtDTFNWd2NibzZfRUNGZmktWXdZZHIzQUVzdzE%3D",
        "time_in_seconds": 32.585,
    },
    {
        "author": {"id": "UCA_ASTgW7mJy7u79oh3GjWQ", "name": "Angel G"},
        "message_type": "text_message",
        "message": "LETS GOOOOOO",
        "message_id": "CjoKGkNKbnA1NG5yNl9FQ0ZSUWRmUW9kU0dBQVFBEhxDTFNEdnFycTZfRUNGWkNyeEFvZHdpUUp2dy0z",
        "time_in_seconds": 32.881,
    },
    {
        "author": {"id": "UCaMEzwyR-ZRSB6_8Pn4dPow", "name": "themoobz"},
        "message_type": "text_message",
        "message": "kon :_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNPV3M3WW5yNl9FQ0ZlQU5yUVlkMTJrS1FBEhtDT3JOMk1UcTZfRUNGV241T0FZZGtBQU8tUTA%3D",
        "time_in_seconds": 33.018,
    },
    {
        "author": {"id": "UCm7LKZTa5yps-Bdq9wt5JTA", "name": "Athreides"},
        "message_type": "text_message",
        "message": "Towa-sama Te Amo",
        "message_id": "CkUKGkNKWGstSW5yNl9FQ0ZZVU9aUW9kSno0TmZ3EidDTHEzcVlUcjZfRUNGWlFTc3dBZG9fNEJBUTE2MjY1ODQ0MDA4MTU%3D",
        "time_in_seconds": 33.266,
    },
    {
        "author": {"id": "UC4jDvtngZKECpncKxp-NkmQ", "name": "vivianpigeon"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjoKGkNJUzVfWW5yNl9FQ0ZkZ0FyUVlkdVBRRlFREhxDTGpzajRicjZfRUNGYzB3S2dvZDFVY0JVQS0w",
        "time_in_seconds": 33.318,
    },
    {
        "author": {"id": "UCBJSteJ5LvA8mxoRj0hb_QQ", "name": "avie"},
        "message_type": "text_message",
        "message": ":_トワ様いえーい::_トワ様いえーい:",
        "message_id": "CkUKGkNQX3VuWXJyNl9FQ0ZWb1ByUVlkVFg4RHVBEidDS0xmNjREcjZfRUNGYWZhY3dFZHhpc0daUTE2MjY1ODQzOTc1NDQ%3D",
        "time_in_seconds": 33.83,
    },
    {
        "author": {"id": "UC6kXeTikSiGiZrTzQEfZZfg", "name": "ラベル"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNOZjd6SXJyNl9FQ0ZaTTFyUVlkRTVvRnRBEidDT3VENDREcjZfRUNGZGR0V0FvZHNxQUQ0UTE2MjY1ODQzOTg1MDA%3D",
        "time_in_seconds": 34.635,
    },
    {
        "author": {"id": "UCLeLH-PQA2ADFOW1AorGUZw", "name": "Hannah Althea"},
        "message_type": "text_message",
        "message": "Konyappi!",
        "message_id": "CjkKGkNLSEU0SXJyNl9FQ0ZWb1ByUVlkVFg4RHVBEhtDTWI3OF96cTZfRUNGVU5IaFFvZFRBVUdydzE%3D",
        "time_in_seconds": 34.956,
    },
    {
        "author": {"id": "UCTSkltVDCgsx_Ct4nG7BFOQ", "name": "UBIKユービック"},
        "message_type": "text_message",
        "message": "こんやっぴー",
        "message_id": "CjkKGkNMQ0M2b3JyNl9FQ0ZaRU5yUVlkQ0I4UEpnEhtDSzdKdFBybTZfRUNGVkpPandvZDdRVUtQdzA%3D",
        "time_in_seconds": 35.054,
    },
    {
        "author": {"id": "UClCtIGFA3V_Tl8jZ8Qd7ZQg", "name": "Dif. Nurina"},
        "message_type": "text_message",
        "message": "kon...:_トワ様いえーい::_トワ様いえーい:",
        "message_id": "CjkKGkNMU2Nfb3JyNl9FQ0ZWb1ByUVlkVFg4RHVBEhtDS3JBNWRyazZfRUNGUTdLY3dFZDY3VU45dzA%3D",
        "time_in_seconds": 35.309,
    },
    {
        "author": {"id": "UCaOs981VIHeNUg6l27mza_w", "name": "Chatsubo"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjkKGkNPUHdqWXZyNl9FQ0ZkZ0FyUVlkdVBRRlFREhtDSUgzeTY3VDZfRUNGY0VsV0FvZEt6NFA0UTE%3D",
        "time_in_seconds": 35.606,
    },
    {
        "author": {"id": "UCtSNv6VgO9Wx7YN8yqlbnMg", "name": "miha 398"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CjoKGkNKemtuSXZyNl9FQ0ZkTVE1d29kZFdVTnBREhxDSmItNGVYcTZfRUNGUUx4T0FZZEhINE1CQS0z",
        "time_in_seconds": 35.913,
    },
    {
        "author": {"id": "UCFJ4GmJv2ZR3jk9bE8yZ8Zg", "name": "はしよし"},
        "message_type": "text_message",
        "message": "こん:_やっぴー:",
        "message_id": "CjoKGkNJZm50WXZyNl9FQ0ZSd2RyUVlkb25FQnR3EhxDT3pEdlBucTZfRUNGZmRDblFrZDlPSUZwdy0w",
        "time_in_seconds": 36.302,
    },
    {
        "author": {"id": "UCH7XtmT0tC-C9RWitB3iemQ", "name": "呵呵"},
        "message_type": "text_message",
        "message": "hi hi",
        "message_id": "CkUKGkNMbkwwSXZyNl9FQ0ZkZ0FyUVlkdVBRRlFREidDTzNFaFlIcjZfRUNGY1IxWUFvZDh0SUpfZzE2MjY1ODQ0MDEwMDU%3D",
        "time_in_seconds": 37.058,
    },
    {
        "author": {"id": "UCE_wvQWLW4ObwEt0-xDBdwg", "name": "Revi"},
        "message_type": "text_message",
        "message": "konyappi",
        "message_id": "CkUKGkNMLTg2WXZyNl9FQ0ZkZ0FyUVlkdVBRRlFREidDTWVVNjRicjZfRUNGVkZBOVFVZDhXa0IzdzE2MjY1ODQ0MDA3MjA%3D",
        "time_in_seconds": 37.134,
    },
    {
        "author": {"id": "UC7Cx27KiMdKtoFIEpC6mRsA", "name": "In9lp"},
        "message_type": "text_message",
        "message": ":_トワ様いえーい:",
        "message_id": "CkUKGkNNcl85b3ZyNl9FQ0ZSYml3UW9kUDAwRy13EidDTmJxdG9mcjZfRUNGYWVrVVFvZEVvMEdIQTE2MjY1ODQ0MDIzMDI%3D",
        "time_in_seconds": 37.502,
    },
    {
        "author": {"id": "UCNDoIL08saEOTZqshRDbOwQ", "name": "ゆきにゃん"},
        "message_type": "text_message",
        "message": "こんやっぴ〜",
        "message_id": "CjoKGkNNT2tqSXpyNl9FQ0ZkZ0FyUVlkdVBRRlFREhxDT3pNLV9fcTZfRUNGYWRDaFFvZFBqd01Xdy0x",
        "time_in_seconds": 37.76,
    },
    {
        "author": {"id": "UChqeFbRtb6abcTRYIY9NBmg", "name": "Mario Olvera"},
        "message_type": "text_message",
        "message": "Let's go my Towa-sama",
        "message_id": "CkUKGkNQVFNsSXpyNl9FQ0ZZVVo1d29keGl3SlBREidDUE8yX0licjZfRUNGYjZuNVFjZGFPY0lhZzE2MjY1ODQzOTcyOTQ%3D",
        "time_in_seconds": 37.941,
    },
    {
        "author": {"id": "UCa2LRcxVk0sQRvMei5ssc3g", "name": "佑偉"},
        "message_type": "text_message",
        "message": "こん:_やっぴー::_トワ様いえーい:",
        "message_id": "CkUKGkNJTGpzSXpyNl9FQ0ZkZ0FyUVlkdVBRRlFREidDT3JXMDRIcjZfRUNGZkZBOVFVZHNBME55ZzE2MjY1ODQ0MDQwNzM%3D",
        "time_in_seconds": 38.242,
    },
    {
        "author": {"id": "UCmkvlGCNxsOKYfziLdC4elw", "name": "SerackSK"},
        "message_type": "text_message",
        "message": "Kon…",
        "message_id": "CjkKGkNOTGdyWXpyNl9FQ0ZRcGI3UW9kLTJRRkNREhtDTVBadE12cTZfRUNGVDdCRmdrZHJnRUJZUTA%3D",
        "time_in_seconds": 38.381,
    },
    {
        "author": {"id": "UC0CE_L15OAYiOwRgm7cUA-A", "name": "TMT"},
        "message_type": "text_message",
        "message": "YOSHA",
        "message_id": "CjoKGkNQTFcwWXpyNl9FQ0ZSd2RyUVlkb25FQnR3EhxDT09MM29UcjZfRUNGUjB6dHdBZHhPUUdhdy0w",
        "time_in_seconds": 38.849,
    },
    {
        "author": {"id": "UC7HlfysV2CqxKvjd6d5NxCw", "name": "Haru [春]"},
        "message_type": "text_message",
        "message": "こん...",
        "message_id": "CkUKGkNJS0wyWXpyNl9FQ0ZZRVhmUW9kWHY4SU53EidDTzZveUlmcjZfRUNGVGZaVEFJZHlEQUxYdzE2MjY1ODQ0MDIxODM%3D",
        "time_in_seconds": 38.978,
    },
    {
        "author": {"id": "UCuPkyqDfoIHYl-sSfhdDtDw", "name": "Yuzuchu"},
        "message_type": "text_message",
        "message": "こんやっぴ~",
        "message_id": "CkUKGkNQYU01b3pyNl9FQ0ZSd2RyUVlkb25FQnR3EidDT3ZQc29mcjZfRUNGUmlteEFvZDdPQU9ZdzE2MjY1ODQ0MDQ0NjU%3D",
        "time_in_seconds": 39.347,
    },
    {
        "author": {"id": "UCidkF_9XheO7n4Bq8FsSwbw", "name": "Hios_Ai"},
        "message_type": "text_message",
        "message": "konyappi~~~",
        "message_id": "CkUKGkNJUzg2b3pyNl9FQ0ZaTTFyUVlkRTVvRnRBEidDT2Y0NW9icjZfRUNGUXc3V0FvZEpMd0JYdzE2MjY1ODQ0MDMzNzk%3D",
        "time_in_seconds": 39.352,
    },
]

if __name__ == "__main__":
    unittest.main()



import logging
import os
from datetime import datetime

def create_logger(
    name,
    folder_path,
    file_name=None,
    sid='Undefined',
    format=None,
    mode="a",
    def_level=logging.ERROR,
    level=logging.DEBUG,
):
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    
    format = format or f"%(asctime)s:%(module)s[%(lineno)d]:%(levelname)s:{f'[{sid}]:' if sid else ''}%(message)s"
    file_name = file_name or get_logname()

    if folder_path:
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        logging.basicConfig(
            level=def_level,
            filename=os.path.join(folder_path, file_name),
            format=format,
            encoding="utf-8",
            filemode=mode,
        )
    else:
        # not create a log file
        logging.basicConfig(
            level=def_level,
            format=format,
            encoding="utf-8",
        )
    console = logging.StreamHandler()
    console.setLevel(def_level)
    console.setFormatter(logging.Formatter(format))
    logging.getLogger(name).addHandler(console)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    return logger


def get_logname() -> str:
    """Gets log name in Y-M-Wn format where n is week number, starts from 0
    Example: 2021-06-W0"""
    weekno = (
        datetime.today().isocalendar()[1]
        - datetime.today().replace(day=1).isocalendar()[1]
    )
    return datetime.today().strftime("%Y-%m-W") + str(weekno) + ".log"



from colorama import Fore
import streamanalyser as sa
import os

# Loading custom context
if __name__ == "__main__":
    analyser_obj = sa.StreamAnalyser(
        "ShB4Wen_HBg",
        verbose=True,
        not_cache=True,
        disable_logs=True,
        msglimit=1000,
        default_context_path=None # set default context path to None to disable premade default contexts
    )
    with analyser_obj as analyser:
        # context source paths should be absolute paths
        analyser.context_source.add(os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "contexts_example.json"
        ))

        # context sources should be added before the analyse function
        analyser.analyse()

        analyser.get_highlights(output_mode="detailed")



import streamanalyser as sa

# Find and open top highlight in the browser
if __name__ == "__main__":
    analyser = sa.StreamAnalyser(
        "pVRvx4FBEwU", msglimit=1000, verbose=True, disable_logs=True, not_cache=True
    )

    with analyser:
        analyser.analyse()
        analyser.get_highlights(top=1)[0].open_in_browser("chrome")



import streamanalyser as sa


# Find users that typed more than one comment
if __name__ == "__main__":
    analyser = sa.StreamAnalyser(
        "pVRvx4FBEwU", msglimit=1000, verbose=True, disable_logs=True, not_cache=True
    )

    with analyser:
        analyser.collect_data()
        analyser.read_data()
        analyser.refine_data()

        for author in analyser.authors:
            # Should use id instead of username for this example,
            # as there might be users with the same name.
            msg_count = len(analyser.find_user_messages(id=author.id))
            if msg_count > 1:
                print(author.name, msg_count)



class DifferentListSizeError(Exception):
    """Raises when list sizes are not the same"""

    pass


class ConstantsNotAscendingError(Exception):
    """Raises when a series of numbers are not in ascending order"""

    pass


class ConstantsNotUniqueError(Exception):
    """Raises when a series of numbers are not unique"""

    pass


class StreamIsLiveOrUpcomingError(Exception):
    """Raises when a live stream is live or upcoming"""

    pass


class DuplicateContextException(Exception):
    """Raised when a context is duplicated"""
    def __init__(self, message, encounters:list):
        self.message = message
        self.encounters = encounters 
    

class PathAlreadyExistsException(Exception):
    """Raised when a path is already exists in the sources"""
    
    pass

class UnexpectedException(Exception):
    """Raised when an unexpected event occurs"""
    
    pass

class ContextsAllCorruptException(Exception):
    """Raised when an all the contexts are corrupt"""
    
    pass